import re
from typing import Dict, List, Tuple

from ._base import split_data_list
from .default_data import DEFAULT_FIELDS_LIST, FIELD_FORMAT_FLAG


class StandardizeEntryBlock(object):
    """Stanndardize entry block.

    Args:
        default_additional_field_list (List[str] = []): Additional default fields.

    Attributes:
        default_fields_list (List[str]): Default fields.
    """

    def __init__(self, default_additional_field_list: List[str] = []) -> None:
        default_fields_old = [d.lower().strip() for d in DEFAULT_FIELDS_LIST]
        default_fields_new = [d.lower().strip() for d in default_additional_field_list]
        self.default_fields_list = list(set(default_fields_old).union(set(default_fields_new)))

    def standardize(self, block: List[str]) -> Tuple[List[str], List[List[str]]]:
        # obtain braces or quotes
        implicit_comments = []
        pre, post = EntryBase().obtain_braces_or_quotes(block)
        if (len(pre) == 0) or (len(post) == 0):
            message = f"Obtain braces or quotes: No standard `pre - {pre}` and `post - {post}`"
            implicit_comments.append(["".join(block), message])
            return [], implicit_comments

        # Obtain fields
        # Support for abbreviations
        field_list = EntryBase().obtain_fields(block, self.default_fields_list)
        if len(field_list) == 0:
            message = "`Obtain fields`: No fields found"
            implicit_comments.append(["".join(block), message])
            return [], implicit_comments

        # Split according to the field pattern
        # Not support abbreviations: ['year = {2019}, journal = ECJ,']
        # TODO Support abbreviations
        pattern = r"\b((?:{})\s*=\s*(?:{}))".format("|".join(field_list), "|".join([rf"{pre}"]))  # compulsory
        block = SplitEntry().split_fields(pattern, block)

        # Append according to the field pattern
        # Support abbreviations
        block = AppendEntry().append_field(field_list, (pre, post), block)

        # Extract
        # Support abbreviations
        block, redundant_list = ExtractEntry().extract(field_list, (pre, post), block)
        if len(redundant_list) != 0:
            message = "`Extract`: Redundant content"
            implicit_comments.append(["".join(redundant_list), message])

        # Check
        # Support abbreviations
        error_dict, block, is_standard_bib_flag = CheckEntry().check(field_list, (pre, post), block)
        if len(error_dict) != 0:
            for key in error_dict:
                implicit_comments.append(["".join(error_dict[key]), f"`Check`: {key}"])

        if not is_standard_bib_flag:
            implicit_comments.append(["".join(block), "`Check`: Not standard bib"])
            return [], implicit_comments

        return block, implicit_comments


class EntryBase(object):
    def __init__(self) -> None:
        pass

    @staticmethod
    def obtain_braces_or_quotes(block: List[str]) -> Tuple[str, str]:
        """Obtain braces or quotes in block.

        Args:
            block (List[str]): block.

        Returns:
            Tuple[str, str]: the tuple of braces or quotes.
        """
        content = "".join(block)
        regex_list = [
            re.compile(r'\btitles*\s*=\s*([{"])', flags=re.I),
            re.compile(r'\bauthors*\s*=\s*([{"])', flags=re.I),
            re.compile(r'\byears*\s*=\s*([{"])', flags=re.I),
            re.compile(r'\bpages*\s*=\s*([{"])', flags=re.I),
            re.compile(r'\burls*\s*=\s*([{"])', flags=re.I),
        ]
        flag_list_list = [sorted(list(set(regex.findall(content)))) for regex in regex_list]

        flag_list_list = [f for f in flag_list_list if len(f) != 0]
        len_list = [len(f) for f in flag_list_list]

        # 0 or 1 or 2 flags
        if (len(len_list) == 0) or (2 in len_list) or (not all([f == flag_list_list[0] for f in flag_list_list])):
            return "", ""

        if flag_list_list[0][0] == "{":
            return "{", "}"
        else:
            return '"', '"'

    def obtain_fields(
        self,
        block: List[str],
        default_fields_list: List[str],
        field_pattern: str = r'[\w\-]+'
    ) -> List[str]:
        r"""Obtain fileds in block.

        Args:
            block (List[str]): block.
            field_pattern (str = r'[\w\-]+'): field pattern.

        Returns:
            List[str]: field list.
        """
        regex = re.compile(rf'({field_pattern})\s*=\s*(?:{"|".join(FIELD_FORMAT_FLAG)})')  # support for abbreviation
        obtain_field_list = list(set(regex.findall("".join(block))))
        obtain_field_list = [field for field in obtain_field_list if field.lower() in default_fields_list]
        return sorted(obtain_field_list)


class SplitEntry(object):
    def __init__(self) -> None:
        super().__init__()

    def split_fields(self, field_pattern: str, block: List[str], last_next: str = "next") -> List[str]:
        return split_data_list(field_pattern, block, last_next)


class AppendEntry(object):
    """Append Patch Bib."""

    def __init__(self) -> None:
        pass

    @staticmethod
    def append_field(field_list: List[str], braces_or_quotes: Tuple[str, str], block: List[str]) -> List[str]:
        """Append.

        Args:
            field_list (List[str]): Append field list.
            braces_or_quotes (Tuple[str, str]): Brace or quote.
            data_list (List[str]): Data list.

        Returns:
            List[str]: new patch bib after appending.
        """
        pre, _ = braces_or_quotes

        temp = rf'[%\s]*(?:{"|".join(field_list)})'
        regex_field = re.compile(rf"{temp}\s*=\s*{pre}", flags=re.I)
        regex_field_abbr = re.compile(rf"{temp}\s*=\s*\w+[\w\-]*", flags=re.I)  # journal = EJC,
        regex_termination = re.compile(r"\s*@[a-zA-Z]*{", flags=re.I)

        # strip and append
        line_index, len_data, new_block = 0, len(block), []
        while line_index < len_data:
            line = block[line_index]
            line_index += 1
            if regex_field.match(line) or regex_termination.match(line) or regex_field_abbr.match(line):
                new_line = line
                while line_index < len_data:
                    line = block[line_index]
                    if regex_field.match(line) or regex_termination.match(line) or regex_field_abbr.match(line):
                        break
                    else:
                        if line.lstrip():
                            new_line = new_line.rstrip() + " " + line.lstrip()  # append
                        line_index += 1
                new_block.append(new_line)
            else:
                new_block.append(line)
        return new_block


class ExtractEntry(object):
    def __init__(self) -> None:
        pass

    def extract(
        self, field_list: List[str], brace_or_quote: Tuple[str, str], block: List[str]
    ) -> Tuple[List[str], List[str]]:
        """Extract.

        Args:
            field_list (List[str]): field list
            brace_or_quote (Tuple[str, str]): (", ") or ({, })
            block (List[str]): the block

        Return:
            Tuple[List[str], List[str]]: main block, redundant part
        """
        pre, post = brace_or_quote

        temp = rf'[%\s]*(?:{"|".join(field_list)})'
        regex_field_two = re.compile(rf"({temp}\s*=\s*{pre})(.*)(\n*)", flags=re.I)
        regex_field_one = re.compile(rf"({temp}\s*=\s*{pre}.*{post})(.*)(\n*)", flags=re.I)
        regex_field_abbr = re.compile(rf"({temp}\s*=\s*\w+[\w\-]*)(.*)(\n*)", flags=re.I)
        regex_termination = re.compile(r"(\s*@[a-zA-Z]*{\s*[\w\-:/\\.\']*)(.*)(\n*)", flags=re.I)

        main_list, redundant_list = [], []

        for line in block:
            new_line, redundant = "", ""
            if mch := regex_termination.match(line):
                one, two, three = mch.groups()
                new_line = one + ",\n"
                if re.sub(r"[\s,\n\}]+", "", two):
                    redundant = two + three

            elif mch := regex_field_abbr.match(line):
                one, two, three = mch.groups()
                new_line = one + ",\n"
                if re.sub(r"[\s,\n\}]+", "", two):
                    redundant = two + three

            elif mch := regex_field_one.match(line):
                one, two, three = mch.groups()
                new_line = self._resub_brace_or_quote(pre, post, one + ",\n")
                if re.sub(r"[\s,\n\}]+", "", two):
                    redundant = two + three

            elif mch := regex_field_two.match(line):
                one, two, three = mch.groups()
                new_line = self._resub_brace_or_quote(pre, post, one + two.strip() + post + ",\n")

            elif line.strip() == "}":
                pass

            else:
                return [], block

            if new_line:
                main_list.append(new_line)
            if redundant:
                redundant_list.append(redundant)

        # for enclosing "@[a-zA-Z]{"
        if main_list:
            main_list.append("}\n")
        return main_list, redundant_list

    def _resub_brace_or_quote(self, pre, post, line: str) -> str:
        if post == "}":
            if line.count(post) > line.count(pre):
                line = re.sub(r'(}[}\s\n,]*)$', '},\n', line)
                line = add_brace_or_quote(pre, post, line)

        elif post == '"':
            if line.count(post) > line.count(pre):
                line = re.sub(r'("["\s\n,]*)$', '",\n', line)
                line = add_brace_or_quote(pre, post, line)
        return line


def add_brace_or_quote(pre, post, line: str):
    if (cpre := line.count(pre)) != (cpost := line.count(post)):
        line_list = list(line)
        if cpre > cpost:
            line_list = line_list[::-1]
            line_list.insert(line_list.index(post), post * (cpre - cpost))
            line_list = line_list[::-1]
        else:
            line_list.insert(line.index(pre), pre * (cpost - cpre))

        line = "".join(line_list)
    return line


class CheckEntry(object):
    @staticmethod
    def check(
        field_list: List[str], brace_or_quote: Tuple[str, str], block: List[str]
    ) -> Tuple[Dict[str, List[str]], List[str], bool]:
        """Check."""
        pre, post = brace_or_quote

        regex_entry = re.compile(r"\s*@[a-zA-Z]+{")
        regex_field = re.compile(fr'\s*(?:{"|".join(field_list)})' + r"\s*=")
        entry_flag, brace_flag = False, False  # minimal conditions
        error_dict: Dict[str, List[str]] = {}
        new_block = []
        for line in block:
            if regex_entry.match(line) and (not entry_flag):  # just iff exsiting one time in single patch bib
                if (line.count("{") != 1) or (line.count(",") != 1):
                    error_dict.setdefault("Failed entry_type", []).append(line)
                else:
                    entry_flag = True
                    new_block.append(line)

            elif regex_field.match(line):
                new_block.append(add_brace_or_quote(pre, post, line))

            elif (line.strip() == "}") and (not brace_flag):  # just iff exsiting one time in single patch bib
                brace_flag = True
                new_block.append(line)

            else:
                error_dict.setdefault("Redundant content`", []).append(line)
        return error_dict, new_block, entry_flag and brace_flag


if __name__ == "__main__":
    pass



"""Initialization.

This submodule incorporates modified source code from the python-bibtexparser project
(https://github.com/sciunto-org/python-bibtexparser), which is licensed under the MIT License.
The original copyright notice and license terms have been preserved in accordance with the license requirements.
"""

__all__ = [
    "Block",
    "Field",
    "Entry",
    "ImplicitComment",
    "ExplicitComment",
    "String",
    "Preamble",
    "ParsingFailedBlock",
    "DuplicateBlockKeyBlock",
    "Library",

    "MiddlewaresStrToStr",
    "MiddlewaresStrToLibrary",
    "MiddlewaresLibraryToLibrary",
    "MiddlewaresLibraryToStr",

    "Splitter",

    "BibtexFormat",
]

from .bibtex_format import BibtexFormat
from .library import Library
from .middlewares_library_to_library import MiddlewaresLibraryToLibrary
from .middlewares_library_to_str import MiddlewaresLibraryToStr
from .middlewares_str_to_library import MiddlewaresStrToLibrary
from .middlewares_str_to_str import MiddlewaresStrToStr
from .model import (
    Block,
    DuplicateBlockKeyBlock,
    Entry,
    ExplicitComment,
    Field,
    ImplicitComment,
    ParsingFailedBlock,
    Preamble,
    String,
)
from .splitter import Splitter



from typing import List, Tuple

from .standardize.do_on_bib import ObtainMarkBlocksDict, SplitBibAccordingToMark
from .standardize.do_on_comment_block import StandardizeCommentBlock
from .standardize.do_on_entry_block import StandardizeEntryBlock
from .standardize.do_on_preamble_block import StandardizePreambleBlock
from .standardize.do_on_string_block import StandardizeStringBlock

MARKS_FLAGS = [
    ["comment", "comment", "C"],        # comment
    ["string", "string", "S"],          # string
    ["preamble", "preamble", "P"],      # preamble
    ["article", "entry", "J"],          # entry
    ["inproceedings", "entry", "C"],    # entry
    ["proceedings", "entry", "B"],      # entry
    ["book", "entry", "B"],             # entry
    ["incollection", "entry", "BS"],    # entry
    ["misc", "entry", "D"],             # entry
    ["unpublished", "entry", "M"],      # entry
    ["techreport", "entry", "R"],       # entry
    ["phdthesis", "entry", "T_D"],      # entry
    ["mastersthesis", "entry", "T_M"],  # entry
]


class StandardizeBib(object):
    """Stanndardize bib.

    Args:
        default_additional_field_list (List[str] = []): Additional default fields.
    """

    def __init__(self, default_additional_field_list: List[str] = []) -> None:
        self._standardize_comment_block = StandardizeCommentBlock()
        self._standardize_entry_block = StandardizeEntryBlock(default_additional_field_list)
        self._standardize_preamble_block = StandardizePreambleBlock()
        self._standardize_string_block = StandardizeStringBlock()

    def standardize(self, data_list: List[str]) -> Tuple[List[str], List[List[str]]]:
        """Generate standard bib.

        Args:
            data_list (List[str]): Bib data.

        Returns:
            List[str]: Standard bib.
        """
        # Initialize
        data_list = "".join(data_list).splitlines(keepends=True)
        data_list = [line for line in data_list if line.strip()]

        # Split data according to mark pattern
        data_list = SplitBibAccordingToMark().split_marks(data_list)

        new_data_list: List[str] = []
        implicit_comment_list: List[List[str]] = []

        # Generate dict
        mark_blocks_dict, temp_implicit_comment_list = ObtainMarkBlocksDict().obtain_dict(data_list, True)
        implicit_comment_list.extend(temp_implicit_comment_list)

        marks, flags = [i[0] for i in MARKS_FLAGS], [i[1] for i in MARKS_FLAGS]
        if not_in := {k: v for k, v in mark_blocks_dict.items() if k not in marks}:
            print(f"Warning: Not standard parts - {not_in}")

        for mark in mark_blocks_dict:
            if mark in marks:
                flag = flags[marks.index(mark)]

                for block in mark_blocks_dict[mark]:
                    block, temp = eval(f"self._standardize_{flag}_block.standardize")(block)
                    new_data_list.extend(block)
                    implicit_comment_list.extend(temp)

        return new_data_list, implicit_comment_list



import re
from typing import List, Tuple


class StandardizeStringBlock(object):
    """Stanndardize string block."""

    def __init__(self) -> None:
        pass

    def standardize(self, block: List[str]) -> Tuple[List[str], List[List[str]]]:
        implicit_comments = []

        regex = re.compile(
            r"@string{" + r"\s*([\w]+)\s*=\s*" + r'(["{])' + r"([\w\-\n]+)" + r'(["}])' + r'(.*)(\n*)', re.DOTALL
        )
        if mch := regex.match("".join(block)):
            a, b, c, d, e, f = mch.groups()
            if ((b == '"') and (d == '"')) or ((b == '{') and (d == '}')):
                block = ["@string{" + a + " = " + b + c.replace("\n", " ").strip() + d + "}\n"]

                if e and e.lstrip()[0] == "}":
                    e = e.lstrip()[1:].lstrip()

                if e.strip():
                    implicit_comments = [[e + f, __class__.__name__]]

            else:
                block = []
                implicit_comments = [["".join(block), __class__.__name__]]
        else:
            block = []
            implicit_comments = [["".join(block), __class__.__name__]]
        return block, implicit_comments



import re
from copy import deepcopy
from typing import List

from ...library import Library
from ...model import (
    Block,
    Entry,
    ExplicitComment,
    ImplicitComment,
    ParsingFailedBlock,
    Preamble,
    String,
)
from ..middleware import LibraryMiddleware

DEFAULT_BLOCK_TYPE_ORDER = (ImplicitComment, ExplicitComment, String, Preamble, Entry, ParsingFailedBlock, Block)


class SortBlocksByTypeAndUserSortKeyMiddleware(LibraryMiddleware):
    """Sorts the blocks of a library by type and `User Sort Key`."""

    def __init__(
        self,
        keep_entry_according_cite_keys: List[str] = [],
        sort_entry_according_field_keys: List[str] = ["year", "volume", "number", "month", "pages"],
        sort_entry_according_field_keys_reverse: bool = True,
    ):
        self._verify_all_types_are_block_types(DEFAULT_BLOCK_TYPE_ORDER)
        self.keep_entry_according_cite_keys = keep_entry_according_cite_keys
        self.sort_entry_according_field_keys = sort_entry_according_field_keys
        self.sort_entry_according_field_keys_reverse = sort_entry_according_field_keys_reverse

        # In-place modification is not yet supported, we make this explicit here,
        super().__init__(allow_inplace_modification=False)

    @staticmethod
    def _verify_all_types_are_block_types(sort_order):
        for t in sort_order:
            if not issubclass(t, Block):
                raise ValueError(
                    "Sort order must only contain Block subclasses, " f"but got {str(t)}"
                )

    # docstr-coverage: inherited
    def transform(self, library: Library) -> Library:
        library = super().transform(library)
        blocks = library.blocks

        _blocks = []

        _implicit_comments = [b for b in blocks if isinstance(b, ImplicitComment)]
        _implicit_comments = sorted(_implicit_comments, key=lambda x: len(x.comment), reverse=False)

        _explicit_comments = [b for b in blocks if isinstance(b, ExplicitComment)]
        _explicit_comments = sorted(_explicit_comments, key=lambda x: len(x.comment), reverse=False)

        _strings = sorted(library.strings, key=lambda x: x.key, reverse=False)

        _preambles = sorted(library.preambles, key=lambda x: len(x.value), reverse=False)

        _entries = library.entries
        if self.keep_entry_according_cite_keys:
            _entries = sorted(_entries, key=self._sort_entry_one, reverse=False)
        else:
            _entries = sorted(_entries, key=self._sort_entry_two, reverse=self.sort_entry_according_field_keys_reverse)

        _failed_blocks = deepcopy(library.failed_blocks)
        _failed_blocks = sorted(_failed_blocks, key=lambda x: len(x.__class__.__name__), reverse=False)

        _others = [b for b in blocks if not (
                isinstance(b, ImplicitComment) or
                isinstance(b, ExplicitComment) or
                isinstance(b, String) or
                isinstance(b, Preamble) or
                isinstance(b, Entry) or
                isinstance(b, ParsingFailedBlock)
            )
        ]
        _others = sorted(_others, key=lambda x: len(x.__class__.__name__), reverse=False)

        # order of blocks
        _blocks.extend(_strings)
        _blocks.extend(_entries)
        _blocks.extend(_preambles)
        _blocks.extend(_explicit_comments)
        _blocks.extend(_implicit_comments)
        _blocks.extend(_failed_blocks)
        _blocks.extend(_others)
        return Library(blocks=_blocks)

    # docstr-coverage: inherited
    @classmethod
    def metadata_key(cls) -> str:
        return "sort_blocks_by_type_and_user_sort_key"

    def _sort_entry_one(self, entry):
        if entry.key in self.keep_entry_according_cite_keys:
            return self.keep_entry_according_cite_keys.index(entry.key)
        else:
            return -1

    def _sort_entry_two(self, entry):
        _sorting_index = []
        _sorting_index.append(entry.entry_type)
        for k in self.sort_entry_according_field_keys:
            if k in entry.fields_dict:
                v = entry.fields_dict[k].value
                # for sorting
                if k == "pages":
                    v = v.split("——")[0].split("--")[0].split("–")[0].split("-")[0]
                if k == "month":
                    v = v.split("–")[0].split("-")[0].split("/")[0]
                _sorting_index.append(v)
            else:
                _sorting_index.append("0")
        return self.sort_strings_with_embedded_numbers("_".join(_sorting_index))

    @staticmethod
    def sort_strings_with_embedded_numbers(s: str) -> List[str]:
        re_digits = re.compile(r"(\d+)")
        pieces = re_digits.split(s)
        pieces[1::2] = map(int, pieces[1::2])
        return pieces



import os
import re
from typing import Any, Dict, List, Union

from pyadvtools import IterateSortDict, standard_path, write_list

from ...bib.bibtexparser import Entry, Library
from ...main import PandocMdTo, PythonRunBib, PythonWriters
from ...utils.utils import html_head, html_style, html_tail, textarea_header, textarea_tail
from ..experiments_base import generate_standard_publisher_abbr_options_dict


def preparation(
    path_storage: str,
    path_output: str,
    output_basename: str,
    pub_type: str,
    issue_or_month_flag: Union[str, List[str]] = "current_issue",
    year_flag: Union[str, List[str]] = "current_year",
    options: Dict[str, Any] = {},
):
    """Prepare paths and flags for data generation.

    Examples
    --------
    |              | current_issue | last_issue | current_month | last_month | given_month | given_months | all_months |
    |--------------|---------------|------------|---------------|------------|-------------|--------------|------------|
    | current_year |               |            |               |            | "2"         | ["1", "3"]   |            |
    | given_year   | "2020"        |            |               |            |             |              |            |
    | given_years  | ["2025"]      |            |               |            |             |              |            |
    | all_years    |               |            |               |            |             |              |            |

    Returns
    -------
    Tuple[str, str, bool]
        Returns (path_root, path_output, combine_flag)
    """
    # default settings
    path_storage = standard_path(path_storage)
    path_output = standard_path(path_output)

    # "absolute_path" or "relative_path"
    absolute_or_relative_path = options.get("absolute_or_relative_path", "absolute_path")

    # Create path components
    yy = "-".join(year_flag) if isinstance(year_flag, List) else year_flag
    im = "-".join(issue_or_month_flag) if isinstance(issue_or_month_flag, List) else issue_or_month_flag

    if options.get("early_access", False):
        base_path = os.path.join(output_basename, f"{pub_type.title()}_Early_Access", f"{yy}_{im}")
        path_output = os.path.join(path_output + "_Early_Access", f"{yy}_{im}")
    else:
        base_path = os.path.join(output_basename, f"{pub_type.title()}", f"{yy}_{im}")
        path_output = os.path.join(path_output, f"{yy}_{im}")

    path_root = base_path if absolute_or_relative_path == "absolute_path" else ""

    # Determine combine flag
    b = options.get("early_access", False) and (year_flag != "all_years")
    c = year_flag == "current_year"
    c = c and (not isinstance(issue_or_month_flag, list)) and (issue_or_month_flag != "all_months")
    combine_flag = b or c

    return path_root, path_output, combine_flag


def generate_from_bibs_and_write(
    path_storage: str,
    path_output: str,
    output_basename: str,
    pub_type: str,
    generate_or_combine: str,
    year_flag: Union[str, List[str]] = "current_year",
    issue_or_month_flag: Union[str, List[str]] = "current_issue",
    options: Dict[str, Any] = {},
) -> None:
    """Generate or combine data from bibliographies.

    Parameters
    ----------
    path_storage : str
        Path to storage directory
    path_output : str
        Path to output directory
    generate_or_combine : str
        Either "generate_data" or "combine_data"
    year_flag : Union[str, List[str]], optional
        Flag for year selection, by default "current_year"
    issue_or_month_flag : Union[str, List[str]], optional
        Flag for issue/month selection, by default "current_issue"
    options : Dict[str, Any], optional
        Additional options, by default {}
    """
    path_root, path_output, combine_flag = preparation(
        path_storage, path_output, output_basename, pub_type, issue_or_month_flag, year_flag, options
    )

    if generate_or_combine == "generate_data":
        publisher_abbr_dict = generate_standard_publisher_abbr_options_dict(path_storage, options)
        for publisher in publisher_abbr_dict:
            pp = os.path.join(path_output, publisher.lower())

            publisher_html_body = []
            for abbr in publisher_abbr_dict[publisher]:
                print(f"*** Processing {publisher.upper()}: {abbr} ***")
                new_options = publisher_abbr_dict[publisher][abbr]

                # Get bibliography path
                path_abbr = os.path.join(path_storage, publisher.lower(), abbr)
                if isinstance(year_flag, str) and year_flag.isdigit():
                    for root, _, files in os.walk(path_abbr, topdown=True):
                        files = [f for f in files if f.endswith(".bib")]
                        if files := [f for f in files if re.search(f"_{year_flag}.bib", f)]:
                            path_abbr = os.path.join(root, files[0])

                # Generate and process library
                library = generate_given_library(path_abbr, issue_or_month_flag, year_flag, new_options)

                # Generate md, tex, pdf, html
                html_body = generate_md_tex_pdf_html(abbr, library, pp, new_options)
                if combine_flag and html_body:
                    publisher_html_body.extend(html_body + ["\n"])

            # Combine for publisher
            if publisher_html_body:
                html_content = _html_content(publisher_html_body[:-1], publisher)
                write_list(html_content, f"{publisher}_all.html", "w", pp, False)

    elif generate_or_combine == "combine_data":
        # Compulsory
        options["include_abbr_list"] = []
        options["exclude_abbr_list"] = []
        publisher_abbr_dict = generate_standard_publisher_abbr_options_dict(path_storage, options)
        for publisher in publisher_abbr_dict:
            print(f"*** Combining papers for {publisher.upper()} ***")
            pp = os.path.join(path_output, publisher.lower())
            absolute_path = os.path.join(path_root, publisher) if len(path_root) > 0 else ""

            link = [f"# {publisher.upper()}\n\n"]
            for abbr in publisher_abbr_dict[publisher]:
                if os.path.exists(os.path.join(pp, abbr, f"{abbr}.html")):
                    ll = os.path.join(absolute_path, abbr, f"{abbr}.html")
                    link.append(f"- [{abbr}]({ll})\n")

            if combine_flag:
                ll = os.path.join(absolute_path, f"{publisher}_all.html")
                link.insert(1, f"- [All Journals]({ll})\n")

            # Process combined content
            if len(link) > 1:
                write_list(link, f"{publisher}_link.md", "w", pp, False)
                PandocMdTo({}).pandoc_md_to_html(pp, pp, f"{publisher}_link.md", f"{publisher}_link.html", True)

            # Clean up
            for name in ["_link"]:
                if os.path.exists(file := os.path.join(pp, f"{publisher}{name}.md")):
                    os.remove(file)
    return None


def generate_given_library(
    original_data: Union[List[str], str, Library],
    issue_or_month_flag: Union[str, List[str]],
    year_flag: Union[str, List[str]] = "current_year",
    options: Dict[str, Any] = {},
) -> Library:
    """Generate a Library object from input data with given filters.

    Parameters
    ----------
    original_data : Union[List[str], str, Library]
        Input bibliography data
    issue_or_month_flag : Union[str, List[str]]
        Flag for issue/month selection
    year_flag : Union[str, List[str]], optional
        Flag for year selection, by default "current_year"
    options : Dict[str, Any], optional
        Additional options, by default {}

    Returns
    -------
    Library
        Processed library object
    """
    _options = {}
    # convert_str_to_library
    _options["is_standardize_bib"] = False  # default is True
    # middlewares_str_to_library.py
    _options["is_display_implicit_comments"] = False  # default is True

    # convert_library_to_library.py
    _options["choose_abbr_zotero_save"] = "save"  # default is "save"
    # middlewares_library_to_library.py
    _options["generate_entry_cite_keys"] = False  # default is False
    _options["function_common_again"] = False  # default is True
    _options["function_common_again_abbr"] = False  # default is True
    _options["function_common_again_zotero"] = False  # default is True
    _options["function_common_again_save"] = False  # default is True

    # convert_library_to_str.py
    # middlewares_library_to_str.py
    _options["is_sort_entry_fields"] = True  # compulsory
    _options["is_sort_blocks"] = True  # compulsory
    _options["sort_entries_by_field_keys_reverse"] = True  # compulsory

    # convert_str_to_str.py
    _options["default_additional_field_list"] = []
    # middlewares_str_to_str.py
    _options["substitute_in_bib"] = False  # default is True

    _options.update(options)
    _python_bib = PythonRunBib(_options)

    # Generate nested entries dictionary
    entry_type_year_volume_number_month_entry_dict = _python_bib.parse_to_nested_entries_dict(original_data)
    old_dict = entry_type_year_volume_number_month_entry_dict

    # Filter by year_flag
    new_dict = {}
    for entry_type in old_dict:
        years = [year for year in old_dict[entry_type]]

        # Update years
        if isinstance(year_flag, List):
            years = sorted(list(set(years).intersection(set(year_flag))))
        elif year_flag.lower().strip() == "all_years":
            years = years
        elif year_flag.lower().strip() == "current_year":
            years = years[:1]
        else:
            if year_flag not in years:
                continue
            else:
                years = [year_flag]

        for year in years:
            new_dict.setdefault(entry_type, {}).update({year: old_dict[entry_type][year]})

    # Filter by issue/month flag
    if issue_or_month_flag in ["current_issue", "last_issue"]:
        return obtain_issue_flag_library(new_dict, issue_or_month_flag)

    return obtain_month_flag_library(new_dict, issue_or_month_flag)


def obtain_issue_flag_library(
    old_dict: Dict[str, Dict[str, Dict[str, Dict[str, Dict[str, List[Entry]]]]]], issue_flag: str = "current_issue"
) -> Library:
    """Filter library by issue flag."""
    old_dict = IterateSortDict(True).dict_update(old_dict)

    entries = []
    for entry_type in old_dict:
        for year in old_dict[entry_type]:
            temp_dict = old_dict[entry_type][year]

            # Article entries
            if entry_type.lower() == "article":
                volumes, numbers, months = [], [], []
                for volume in (volumes := [volume for volume in temp_dict]):
                    for number in (numbers := [number for number in temp_dict[volume]]):
                        months = [month for month in temp_dict[volume][number]]
                        break
                    break

                if issue_flag == "current_issue":  # current volume, current issue, and current month
                    entries.extend(temp_dict[volumes[0]][numbers[0]][months[0]])

                elif issue_flag == "last_issue":
                    # Logic for getting previous issue
                    if len(months) == 1:
                        if len(numbers) == 1:
                            if len(volumes) == 1:
                                entries.extend(temp_dict[volumes[0]][numbers[0]][months[0]])
                            else:
                                numbers = [number for number in temp_dict[volumes[1]]]
                                months = [month for month in temp_dict[volumes[1]][numbers[0]]]
                                entries.extend(temp_dict[volumes[1]][numbers[0]][months[0]])
                        else:
                            months = [month for month in temp_dict[volumes[0]][numbers[1]]]
                            entries.extend(temp_dict[volumes[0]][numbers[1]][months[0]])
                    else:
                        entries.extend(temp_dict[volumes[0]][numbers[0]][months[1]])

                else:
                    print(f"Unknown issue flag: {issue_flag}.")

            else:
                # Non-article entries
                for volume in temp_dict:
                    for number in temp_dict[volume]:
                        for month in temp_dict[volume][number]:
                            entries.extend(temp_dict[volume][number][month])

    return Library(entries)


def obtain_month_flag_library(
    old_dict: Dict[str, Dict[str, Dict[str, Dict[str, Dict[str, List[Entry]]]]]],
    month_flag: Union[str, List[str]] = "current_month",
) -> Library:
    """Filter library by month flag."""
    new_dict = {}
    for entry_type in old_dict:
        for year in old_dict[entry_type]:

            for volume in old_dict[entry_type][year]:
                for number in old_dict[entry_type][year][volume]:
                    for month in old_dict[entry_type][year][volume][number]:
                        new_dict.setdefault(entry_type, {}).setdefault(year, {}).setdefault(month, {}).setdefault(
                            volume, {}
                        ).setdefault(number, []).extend(old_dict[entry_type][year][volume][number][month])

    # Sort
    old_dict = IterateSortDict(True).dict_update(new_dict)

    entries = []
    for entry_type in old_dict:
        for year in old_dict[entry_type]:
            temp_dict = old_dict[entry_type][year]
            default_months = [month for month in temp_dict]

            # Update month
            new_months = []
            if month_flag == "current_month":  # current_month
                new_months = [default_months[0]]
            elif month_flag == "last_month":  # last_month
                new_months = [default_months[1]] if len(default_months) > 1 else []
            elif month_flag == "all_months":  # all months
                new_months = default_months
            else:
                if isinstance(month_flag, str):  # given month
                    if month_flag in default_months:
                        new_months = [month_flag]
                else:
                    for month in month_flag:  # given months
                        if month in default_months:
                            new_months.append(month)

            for month in new_months:
                for volume in temp_dict[month]:
                    for number in temp_dict[month][volume]:
                        entries.extend(temp_dict[month][volume][number])

    return Library(entries)


def generate_md_tex_pdf_html(
    abbr_standard: str,
    original_bib_data: Union[List[str], str, Library],
    path_output: str,
    options: Dict[str, Any] = {},
) -> List[str]:
    """Generate markdown and LaTeX from bibliography data."""
    options_: dict = {
        # convert_str_to_library
        "is_standardize_bib": False,
        # middlewares_str_to_library.py
        "is_display_implicit_comments": False,
        #
        # convert_library_to_library.py
        # middlewares_library_to_library.py
        "function_common_again": False,
        "function_common_again_abbr": False,
        "function_common_again_zotero": False,
        "function_common_again_save": False,
        "abbr_index_article_for_abbr": 2,
        "abbr_index_inproceedings_for_abbr": 2,
        #
        # convert_library_to_str.py
        "empty_entry_cite_keys": True,
        # middlewares_library_to_str.py
        "is_sort_entry_fields": True,
        "is_sort_blocks": True,
        "sort_entries_by_field_keys_reverse": True,
    }
    options_.update(options)

    # Process bibliography data
    _python_bib = PythonRunBib(options_)
    _, zotero_library, _ = _python_bib.parse_to_multi_standard_library(original_bib_data)

    _python_writer = PythonWriters(options_)

    # Generate HTML content body
    html_body = []
    for entry in zotero_library.entries:
        html_body.append(_format_entry(entry, abbr_standard, _python_writer.write_to_str([entry])))

    if len(html_body) > 0:
        html_body = (
            [f'<h2 id="{abbr_standard.lower()}">{abbr_standard} - {len(zotero_library.entries)}</h2>\n', "<ul>\n"]
            + html_body
            + ["</ul>\n"]
        )

        html_content = _html_content(html_body, abbr_standard)

        # Write output file
        write_list(html_content, f"{abbr_standard}.html", "w", os.path.join(path_output, abbr_standard), False)

    return html_body


def _html_content(html_body, abbr_standard):
    html_content = [html_head.format(abbr_standard), html_style, "\n"]
    html_content.extend(html_body)
    html_content.extend([html_tail])
    return html_content


def _format_entry(entry, abbr, data_list):
    """Format a single bibliography entry into HTML."""
    number = entry["number"] if "number" in entry else ""
    pages = entry["pages"] if "pages" in entry else ""
    title = entry["title"] if "title" in entry else ""
    year = entry["year"] if "year" in entry else ""
    volume = entry["volume"] if "volume" in entry else ""

    url = ""
    if "doi" in entry:
        doi = entry["doi"]
        url = doi if doi.startswith("http") else f"https://doi.org/{doi}"
    elif "url" in entry:
        url = entry["url"]

    line = _format_entry_apa(title, year, volume, number, pages, url, abbr)

    line = f"<li><details>\n<summary>\n{line.strip()}\n</summary>\n"

    return line + textarea_header + "".join(data_list).rstrip() + textarea_tail + "\n</details></li>\n"


def _format_entry_apa(title, year, volume, number, pages, url, abbr):
    line = f"({year}). {title}. <em>{abbr}</em>"
    if volume:
        line += f", <em>{volume}</em>"
        if number:
            line += f"({number})"

    if pages:
        line += f", {pages}"

    line += "."

    if url:
        line += f" (<a href='{url}'>www</a>)"

    return line



from typing import Any, Dict, Tuple

from ..bibtexparser import Library, MiddlewaresLibraryToLibrary


class ConvertLibrayToLibrary(object):
    """Convert library to library.

    Args:
        options (Dict[str, Any]): Options. Default is {}.

    Attributes:
        choose_abbr_zotero_save (str): Choose "abbr", "zotero", or "save". Default is "save".
    """

    def __init__(self, options: Dict[str, Any] = {}) -> None:

        self.choose_abbr_zotero_save = options.get("choose_abbr_zotero_save", "save")

        self._middleware_library_library = MiddlewaresLibraryToLibrary(options)

    def generate_single_library(self, library: Library) -> Library:
        func = eval(f"_x.function_{self.choose_abbr_zotero_save}", {}, {"_x": self._middleware_library_library})
        library = func(library)

        return library

    def generate_multi_libraries(self, library: Library) -> Tuple[Library, Library, Library]:
        abbr_library, zotero_library, save_library = self._middleware_library_library.functions(library)

        return abbr_library, zotero_library, save_library



import copy
import re
from typing import Dict, List, Tuple

from ...bib.bibtexparser import Library
from ...main import PythonRunBib, PythonWriters
from .search_writers import WriteInitialResult, WriteSeparateResult


def search_keywords_core(keywords_list_list: List[List[str]], library: Library, field: str) -> Tuple[Library, Library]:
    """Search keywords in `field` such as `title` or `abstract` or `keywords`."""
    search_library = []
    no_search_library = []

    for entry in library.entries:
        flag = False
        content = entry[field] if field in entry else ""
        if content:
            content = re.sub("{", "", content)
            content = re.sub("}", "", content)

            # All keywords from keyword_list_list[0] should be found in bib
            flag = all([re.search(keyword, content, flags=re.I) for keyword in keywords_list_list[0]])
            if flag and (len(keywords_list_list) == 2):
                # Any keywords from keyword_list_list[1] found in bib will results in False flag.
                flag = not any([re.search(keyword, content, flags=re.I) for keyword in keywords_list_list[1]])

        if flag:
            search_library.append(entry)
        else:
            no_search_library.append(entry)

    return Library(search_library), Library(no_search_library)


class SearchInitialResult(object):
    """Search initial result.

    Args:
        options: dict

    Attributes:
        options: dict

        print_on_screen (bool = False): print on screen
        deepcopy_library_for_every_field (bool = False): deepcopy library for every field
    """

    def __init__(self, options: dict) -> None:
        self.options = options

        self.print_on_screen: bool = options.get("print_on_screen", False)
        self.deepcopy_library_for_every_field = options.get("deepcopy_library_for_every_field", False)

        self._python_bib = PythonRunBib(options)

        _options = {}
        _options["empty_entry_cite_keys"] = True
        _options.update(self.options)
        self._python_writer = PythonWriters(_options)

    def main(
        self,
        search_field_list: List[str],
        path_initial: str,
        library: Library,
        keywords_type: str,
        keywords_list_list: List[List[str]],
        combine_keywords: str,
        output_prefix: str,
        path_separate: str,
    ) -> Tuple[List[str], Dict[str, List[List[str]]], Dict[str, int], Library]:
        """Search."""
        error_pandoc_md_md, field_data_dict, no_search_library = [], {}, library
        field_number_dict: Dict[str, int] = {}

        for field in search_field_list:
            if len(no_search_library.entries) == 0:
                continue

            # Search
            search_library, no_search_library = search_keywords_core(keywords_list_list, no_search_library, field)
            field_number_dict.update({field: len(search_library.entries)})

            # Deepcopy library for every field
            if self.deepcopy_library_for_every_field:
                no_search_library = copy.deepcopy(library)

            # Operate on the search library (deepcopy)
            libraries = self._python_bib.parse_to_multi_standard_library(copy.deepcopy(search_library))
            library_for_abbr, library_for_zotero, library_for_save = libraries

            if self.print_on_screen:
                print("".join(self._python_writer.write_to_str(library_for_zotero)))
                continue
            if not (library_for_abbr.entries and library_for_zotero.entries and library_for_save.entries):
                continue

            # Initially write tex, bib, and md files
            data_temp, temp_error_pandoc_md_md = WriteInitialResult(copy.deepcopy(self.options)).main(
                path_initial,
                output_prefix,
                field,
                keywords_type,
                combine_keywords,
                library_for_abbr,
                library_for_zotero,
                library_for_save,
            )

            # Separatelly write with the method 'a' for '_basic', '_beauty', '_complex'
            WriteSeparateResult().main(copy.deepcopy(data_temp), field, keywords_type, combine_keywords, path_separate)

            # Save for combined results
            field_data_dict.update({field: copy.deepcopy(data_temp)})
            error_pandoc_md_md.extend(temp_error_pandoc_md_md)

        return error_pandoc_md_md, field_data_dict, field_number_dict, no_search_library



import os
import re
from typing import Any, Dict, List, Optional, Tuple, Union

from pyadvtools import (
    write_list,
)

from ..bib.bibtexparser import BibtexFormat, Block, Library
from ..bib.core import ConvertLibrayToStr
from .basic_input import BasicInput
from .python_run_bib import PythonRunBib


class PythonWriters(BasicInput):
    """Python writers.

    Args:
        options (Dict[str, Any]): Options.

    Attributes:
        bib_for_abbr_name (str = "abbr.bib"): Bib for abbr name.
        bib_for_zotero_name (str = "zotero.bib"): Bib for zotero name.
        bib_for_save_name (str = "save.bib"): Bib for save name.
        choose_abbr_zotero_save (str = "save"): Choose "abbr", "zotero", or "save".
        display_google_connected_scite (List[str] = ["google", "connected", "scite"]): Choose multiple items.
    """

    def __init__(self, options: Dict[str, Any]) -> None:
        options["is_sort_entry_fields"] = options.get("is_sort_entry_fields", True)
        options["is_sort_blocks"] = options.get("is_sort_blocks", True)
        options["sort_entries_by_field_keys_reverse"] = options.get("sort_entries_by_field_keys_reverse", True)
        super().__init__(options)

        self.bib_for_abbr_name = options.get("bib_for_abbr_name", "abbr.bib")
        self.bib_for_zotero_name = options.get("bib_for_zotero_name", "zotero.bib")
        self.bib_for_save_name = options.get("bib_for_save_name", "save.bib")
        self.choose_abbr_zotero_save = options.get("choose_abbr_zotero_save", "save")

        self.display_google_connected_scite = options.get(
            "display_google_connected_scite", ["google", "connected", "scite"]
        )

        self.bibtex_format_indent = options.get("bibtex_format_indent", "  ")
        self.bibtex_format_trailing_comma = options.get("bibtex_format_trailing_comma", True)
        self.bibtex_format_block_separator = options.get("bibtex_format_block_separator", "")

        bibtex_format = BibtexFormat()
        bibtex_format.indent = self.bibtex_format_indent
        bibtex_format.block_separator = self.bibtex_format_block_separator
        bibtex_format.trailing_comma = self.bibtex_format_trailing_comma
        self.bibtex_format: Optional[BibtexFormat] = bibtex_format

        self._python_bib = PythonRunBib(self.options)

    def write_to_str(self, library: Union[Library, List[Block]]) -> List[str]:
        """Serialize a BibTeX database to a string.

        Args:
            library (Library | List[Block]): BibTeX database to serialize.
            bibtex_format (Optional[BibtexFormat] = None):

        """
        return ConvertLibrayToStr(self.options).generate_str(library, self.bibtex_format)

    def write_to_file(
        self,
        original_data: Union[Library, List[Block], List[str]],
        file_name: str,
        write_flag: str = "w",
        path_storage: Optional[str] = None,
        check: bool = True,
        delete_first_empty: bool = True,
        delete_last_empty: bool = True,
        compulsory: bool = False,
        delete_original_file: bool = False,
    ) -> None:
        """Write.

        Args:
            original_data (Union[Library, List[Block], List[str]]): data
            file_name (str): file name
            write_flag (str = "w"): write flag
            path_storage (Optional[str] = None): path storage
            check (bool = True): check
            delete_first_empty (bool = True): delete first empty
            delete_last_empty (bool = True): delete last empty
            compulsory (bool = False): compulsory
            delete_original_file (bool = False): delete original file
            bibtex_format (Optional[BibtexFormat] = None):

        """
        _options = {}
        _options.update(self.options)
        _library_str = ConvertLibrayToStr(_options)

        if isinstance(original_data, Library):
            data_list = _library_str.generate_str(original_data, self.bibtex_format)
        elif isinstance(original_data, list):
            if all([isinstance(line, str) for line in original_data]):
                data_list = [line for line in original_data if isinstance(line, str)]
            else:
                data_list = [line for line in original_data if isinstance(line, Block)]
                data_list = _library_str.generate_str(data_list, self.bibtex_format)

        write_list(
            data_list,
            file_name,
            write_flag,
            path_storage,
            check,
            delete_first_empty,
            delete_last_empty,
            compulsory,
            delete_original_file,
        )
        return None

    def write_multi_library_to_file(
        self,
        path_output: str,
        bib_for_abbr: Union[Library, List[Block]],
        bib_for_zotero: Union[Library, List[Block]],
        bib_for_save: Union[Library, List[Block]],
        given_cite_keys: List[str] = [],
        **kwargs,
    ) -> Tuple[str, str, str]:
        _options = {}
        _options.update(self.options)
        _options["keep_entries_by_cite_keys"] = given_cite_keys
        _options["sort_entries_by_cite_keys"] = given_cite_keys

        bib_abbr = ConvertLibrayToStr(_options).generate_str(bib_for_abbr, **kwargs)
        write_list(bib_abbr, self.bib_for_abbr_name, "w", path_output, False, **kwargs)

        bib_zotero = ConvertLibrayToStr(_options).generate_str(bib_for_zotero, **kwargs)
        write_list(bib_zotero, self.bib_for_zotero_name, "w", path_output, False, **kwargs)

        bib_save = ConvertLibrayToStr(_options).generate_str(bib_for_save, **kwargs)
        write_list(bib_save, self.bib_for_save_name, "w", path_output, False, **kwargs)

        full_bib_for_abbr = os.path.join(path_output, self.bib_for_abbr_name)
        full_bib_for_zotero = os.path.join(path_output, self.bib_for_zotero_name)
        full_bib_for_save = os.path.join(path_output, self.bib_for_save_name)
        return full_bib_for_abbr, full_bib_for_zotero, full_bib_for_save

    def output_key_url_http_bib_dict(self, library: Library) -> Dict[str, List[List[str]]]:
        _options = {}
        _options.update(self.options)
        _options["empty_entry_cite_keys"] = True

        key_url_http_bib_dict: Dict[str, List[List[str]]] = {}

        for key, entry in library.entries_dict.items():

            url = entry["url"] if "url" in entry else ""
            if len(url) == 0:
                url = entry["doi"] if "doi" in entry else ""
                if (len(url) != 0) and (not re.match(r"https*://", url)):
                    url = f"https://doi.org/{url}"

            link_list = self._generate_link_list(entry)
            patch_bib = ConvertLibrayToStr(_options).generate_str([entry])

            v: List[List[str]] = [[], [], patch_bib]

            if len(url) != 0:
                v[0] = [url + "\n"]
                link_list.insert(0, rf"[www]({url})")

            join_link = []
            if link_list:
                for i in range(len(link_list) - 1):
                    join_link.append(link_list[i].strip() + " |\n")
                join_link.append(link_list[-1].strip() + "\n")

                join_link[0] = "(" + join_link[0]
                join_link[-1] = join_link[-1].strip() + ")\n"

                v[1] = join_link

            key_url_http_bib_dict.update({key: v})
        return key_url_http_bib_dict

    def _generate_link_list(self, entry) -> List[str]:
        title = entry["title"] if "title" in entry else ""
        if not title:
            return []

        title = re.sub(r"\s+", "+", title)
        url_google = f"https://scholar.google.com/scholar?q={title}"
        url_connected = f"https://www.connectedpapers.com/search?q={title}"
        url_scite = f"https://scite.ai/search?q={title}"

        # Search cited number
        cited_number = entry["annotation"] if "annotation" in entry else ""
        if cited_number:
            cited_number = re.sub(r"[^0-9]+", "", cited_number)
            cited_number = int(cited_number) if cited_number.isdigit() else ""
            google = f"[Google Scholar: {cited_number}]({url_google})"
        else:
            google = f"[Google Scholar]({url_google})"

        connected = f"[Connected Papers]({url_connected})"
        scite = f"[Scite]({url_scite})"

        link_list = []
        for i, j in zip(["google", "connected", "scite"], [google, connected, scite]):
            if i in self.display_google_connected_scite:
                link_list.append(j)

        return link_list



import os
import re
import subprocess
from typing import Any, Dict, List, Optional

from pyadvtools import (
    delete_files,
    insert_list_in_list,
    read_list,
    write_list,
)

from .basic_input import BasicInput


class PythonRunTex(BasicInput):
    """Python tex.

    Args:
        options (Dict[str, Any]): Options.

    Attributes:
        run_latex (bool = False): Whether to run latex.
        pdflatex_xelatex (str = "pdflatex"): pdflatex or xelatex.
        delete_run_latex_cache (bool = True): Whether to delete run latex cache.
    """

    def __init__(self, options: Dict[str, Any]) -> None:
        super().__init__(options)

        # for tex
        self.final_output_main_tex_name: str = options.get("final_output_main_tex_name", "")
        self.run_latex: bool = options.get("run_latex", False)
        self.pdflatex_xelatex: str = options.get("pdflatex_xelatex", "pdflatex")  # pdflatex, xelatex
        self.delete_run_latex_cache: bool = options.get("delete_run_latex_cache", True)
        self.latex_clean_file_types: Optional[List[str]] = options.get("latex_clean_file_types", None)

    def generate_standard_tex_data_list(
        self,
        data_list_body: List[str],
        output_tex_name: str,
        path_output: str,
        figure_folder_name: str = "figs",
        tex_folder_name: str = "tex",
        bib_folder_name: str = "bib",
        bib_name: str = "abbr.bib",
        template_name: str = "article",
    ) -> List[str]:
        # for figures
        for i in range(len(data_list_body)):
            if re.search(r"\\includegraphics", data_list_body[i]):
                data_list_body[i] = data_list_body[i].replace(f".{os.sep}Figures{os.sep}", f".{os.sep}{figure_folder_name}{os.sep}")
                data_list_body[i] = data_list_body[i].replace(f"Figures{os.sep}", f"{figure_folder_name}{os.sep}")
        write_list(data_list_body, output_tex_name, "w", os.path.join(path_output, tex_folder_name), False)

        self._special_operate_tex(
            data_list_body, template_name, output_tex_name, path_output, tex_folder_name, bib_folder_name, bib_name
        )
        return data_list_body

    # for tex files
    def _special_operate_tex(
        self,
        data_list_body: List[str],
        template_name: str,
        output_tex_name: str,
        path_output: str,
        tex_folder_name: str,
        bib_folder_name: str,
        bib_name: str,
    ) -> None:
        # read template data
        template_h, template_t = [], []
        if template_name.lower() == "paper":
            template_h, template_t = self.article_template_header_tex, self.article_template_tail_tex
        elif template_name.lower() == "beamer":
            template_h, template_t = self.beamer_template_header_tex, self.beamer_template_tail_tex

        # style
        if usepackages := self.usepackages_tex:
            usepackages.insert(0, "\n")

        # command
        if math_commands := self.math_commands_tex:
            math_commands.insert(0, "\n")

        # main name
        main_name = self.final_output_main_tex_name
        if len(main_name) == 0:
            main_name = output_tex_name.split(".text")[0] + "_main.tex"
        if main_name.lower() == output_tex_name.lower():
            main_name = main_name.split(".tex")[0] + "_.tex"
        if main_name[-4:] != ".tex":
            main_name = main_name + ".tex"

        data_list = []
        if (len(template_h) != 0) and (len(template_t) != 0):
            # header
            data_list = insert_list_in_list(template_h, usepackages, r"\\documentclass", "after")
            data_list = insert_list_in_list(data_list, math_commands, r"\\documentclass", "after")

            if template_name.lower() == "beamer":
                data_list = insert_list_in_list(data_list, ["\n\\def\\allfiles{}\n"], r"\\documentclass", "after")

            if self.pdflatex_xelatex == "xelatex":
                data_list = insert_list_in_list(data_list, ["\n\\def\\cn{}\n"], r"\\documentclass", "after")

            # for bib
            regex = re.compile(r"\\addbibresource{" + os.path.join(".", "References", "References.bib") + "}")
            for i in range(len(data_list)):
                mch = regex.search(data_list[i])
                if not mch:
                    continue
                if bib_folder_name:
                    data_list[i] = "\\addbibresource{" + os.path.join(".", bib_folder_name, bib_name) + "}\n"
                else:
                    data_list[i] = "\\addbibresource{" + os.path.join(".", bib_name) + "}\n"

            # body
            if len(data_list_body) != 0:
                data_list.append("\n")
                data_list.extend(data_list_body)
                data_list.append("\n")

            # tail
            data_list.extend(template_t)

            # save file
            write_list(data_list, main_name, "w", path_output, False)
        else:
            data_list = read_list(output_tex_name, "r", os.path.join(path_output, tex_folder_name))
            write_list(data_list, main_name, "w", path_output, False)

        # run latex
        if self.run_latex:
            os.chdir(path_output)
            cmd = "latexmk -{} {}".format(self.pdflatex_xelatex, main_name)
            try:
                subprocess.run(cmd, shell=True, check=True, capture_output=True, text=True)
            except subprocess.CalledProcessError as e:
                print("Error in Run LaTex:", e.stderr)

        # delete cache
        if self.delete_run_latex_cache:
            if self.latex_clean_file_types is not None:
                postfix = self.latex_clean_file_types
            else:
                postfix = ['.aux', '.bbl', '.bcf', '.blg', '.fdb_latexmk', '.fls', '.log', '.out', '.run.xml']
                postfix.extend(['.synctex.gz', '.gz', '.nav', '.snm', '.toc', '.xdv'])
            delete_files(path_output, postfix)
        return None



from ...library import Library
from ...model import Block, Entry
from ..middleware import BlockMiddleware


class NormalizePagesInEntry(BlockMiddleware):
    """Normalize field `pages` of an entry by deleting redundant part or generating when not existed."""

    def __init__(self, allow_inplace_modification: bool = True):
        super().__init__(allow_inplace_modification=allow_inplace_modification, allow_parallel_execution=True)

    # docstr-coverage: inherited
    def transform_entry(self, entry: Entry, library: Library) -> Block:
        if "pages" in entry:
            # 5-10-5-10 -> 5-10
            page_list = []
            for page in entry["pages"].split("-"):  # English hyphen
                for p in page.strip().split("—"):  # Chinese hyphen
                    if p.strip():
                        page_list.append(p.strip())

            page_list = sorted(set(page_list), key=page_list.index)
            entry["pages"] = "-".join(page_list)
        else:
            # pages = {12:1-37}
            if "articleno" in entry and "numpages" in entry:
                entry["pages"] = f'{entry["articleno"]}:1-{entry["numpages"]}'
        return entry



import logging
from typing import Dict, List, Set

from ...library import Library
from ...model import Entry, Field
from ..middleware import BlockMiddleware

logger = logging.getLogger(__name__)


class NormalizeEntryFieldKeys(BlockMiddleware):
    """Normalize field keys to lowercase.

    In case of conflicts (e.g. both 'author' and 'Author' exist in the same entry),
    a warning is emitted, and the last value wins.

    Some other middlewares, such as `SeparateCoAuthors`, assume lowercase key names.
    """

    def __init__(self, allow_inplace_modification: bool = True):
        super().__init__(allow_inplace_modification=allow_inplace_modification, allow_parallel_execution=True)

    # docstr-coverage: inherited
    def transform_entry(self, entry: Entry, library: Library) -> Entry:
        seen_normalized_keys: Set[str] = set()
        new_fields_dict: Dict[str, Field] = {}
        for field in entry.fields:
            normalized_key: str = field.key.lower()
            # if the normalized key is already present, apply "last one wins"
            # otherwise preserve insertion order
            # if a key is overwritten, emit a detailed warning
            # if performance is a concern, we could emit a warning with only {entry.key}
            # to remove "seen_normalized_keys" and this if statement
            if normalized_key in seen_normalized_keys:
                logger.warning(
                    f"NormalizeFieldKeys: in entry '{entry.key}': "
                    + f"duplicate normalized key '{normalized_key}' "
                    + f"(original '{field.key}'); overriding previous value"
                )
            seen_normalized_keys.add(normalized_key)
            field.key = normalized_key
            new_fields_dict[normalized_key] = field

        new_fields: List[Field] = list(new_fields_dict.values())
        entry.fields = new_fields

        return entry



import os
import re
import shutil
from typing import Any, Dict, List, Optional, Tuple, Union

from pyadvtools import (
    combine_content_in_list,
    read_list,
    standard_path,
    write_list,
)

from ..bib.bibtexparser import Library
from ..main import BasicInput, PythonRunBib, PythonRunMd, PythonRunTex, PythonWriters


class PyRunBibMdTex(BasicInput):
    """A class for processing BibTeX, Markdown and LaTeX files with various operations.

    This class provides functionality to handle references, figures, and content conversion
    between Markdown and LaTeX formats.
    """

    def __init__(
        self, path_output: str, tex_md_flag: str = ".md", template_name: str = "paper", options: Dict[str, Any] = {}
    ) -> None:
        """Initialize the PyRunBibMdTex instance.

        Parameters
        ----------
        path_output : str
            Output directory path for processed files
        tex_md_flag : str, optional
            Flag indicating whether to process as LaTeX (".tex") or Markdown (".md"),
            by default ".md"
        template_name : str, optional
            Template type to use ("paper" or "beamer"), by default "paper"
        options : Dict[str, Any], optional
            Additional configuration options, by default {}

        Raises
        ------
        AssertionError
            If tex_md_flag is not ".tex" or ".md"
            If template_name is not "paper" or "beamer"
        """
        super().__init__(options)

        self.tex_md_flag = re.sub(r"\.+", ".", "." + tex_md_flag)
        assert self.tex_md_flag in [".tex", ".md"], f"{tex_md_flag} must be `.tex` or `.md`."
        self.template_name = template_name.lower()
        assert self.template_name in ["paper", "beamer"], f"{template_name} must be `paper` or `beamer`."
        self.path_output = standard_path(path_output)

        # Configuration options
        self.generate_html = options.get("generate_html", False)
        self.generate_tex = options.get("generate_tex", True)
        self.shutil_figures = options.get("shutil_figures", True)

        # Folder name configurations
        self.figure_folder_name = options.get("figure_folder_name", "fig")  # "" or "figs" or "main"
        self.bib_folder_name = options.get("bib_folder_name", "bib")  # "" or "bibs" or "main"
        self.md_folder_name = options.get("md_folder_name", "md")  # "" or "mds" or "main"
        self.tex_folder_name = options.get("tex_folder_name", "tex")  # "" or "texes" or "main"

        # Cleanup options
        self.delete_original_md_in_output_folder = options.get("delete_original_md_in_output_folder", False)
        self.delete_original_tex_in_output_folder = options.get("delete_original_tex_in_output_folder", False)
        self.delete_original_bib_in_output_folder = options.get("delete_original_bib_in_output_folder", False)

        # Initialize helper classes
        self._python_bib = PythonRunBib(self.options)
        self._python_md = PythonRunMd(self.options)
        self._python_tex = PythonRunTex(self.options)
        self._python_writer = PythonWriters(self.options)

    def run_files(
        self, file_list_md_tex: List[str], output_prefix: str = "", output_level: str = "next"
    ) -> Tuple[List[str], List[str]]:
        """Process a list of Markdown or LaTeX files.

        Parameters
        ----------
        file_list_md_tex : List[str]
            List of input file paths (Markdown or LaTeX)
        output_prefix : str, optional
            Prefix for output files, by default ""
        output_level : str, optional
            Output directory level ("previous", "current", or "next"), by default "next"

        Returns
        -------
        Tuple[List[str], List[str]]
            Tuple containing processed Markdown content and LaTeX content
        """
        file_list_md_tex = [f for f in file_list_md_tex if f.endswith(self.tex_md_flag)]
        data_list_list = [read_list(standard_path(f), "r") for f in file_list_md_tex]
        if all([len(data_list) == 0 for data_list in data_list_list]):
            return [], []

        file_base_name = os.path.splitext(os.path.basename(file_list_md_tex[0]))[0]
        output_prefix = output_prefix if output_prefix else file_base_name

        data_list_md_tex = combine_content_in_list(data_list_list, ["\n"])

        content_md, content_tex = self.python_run_bib_md_tex(
            output_prefix, data_list_md_tex, self.path_bibs, output_level
        )
        return content_md, content_tex

    def python_run_bib_md_tex(
        self,
        output_prefix: str,
        data_list_md_tex: List[str],
        original_bib_data: Union[List[str], str, Library],
        output_level: str = "next",
    ) -> Tuple[List[str], List[str]]:
        """Process BibTeX, Markdown and LaTeX content.

        Parameters
        ----------
        output_prefix : str
            Prefix for output files
        data_list_md_tex : List[str]
            List of content lines (Markdown or LaTeX)
        original_bib_data : Union[List[str], str, Library]
            BibTeX data in various formats
        output_level : str, optional
            Output directory level ("previous", "current", or "next"), by default "next"

        Returns
        -------
        Tuple[List[str], List[str]]
            Tuple containing processed Markdown content and LaTeX content
        """
        # Basic file names
        output_tex, output_md = output_prefix + ".tex", output_prefix + ".md"

        if len(data_list_md_tex) == 0:
            original_bib_data = self._python_bib.parse_to_single_standard_library(original_bib_data)
            if not original_bib_data.entries:
                return [], []

            data_list_md_tex = []
            for entry in original_bib_data.entries:
                data_list_md_tex.append(f"- [@{entry.key}]\n\n")
            data_list_md_tex.insert(0, f"## {output_prefix} - {len(data_list_md_tex)}\n\n")

        # Determine output path based on level
        if output_level == "previous":
            path_output = os.path.dirname(self.path_output)
        elif output_level == "next":
            path_output = os.path.join(self.path_output, output_prefix)
        elif output_level == "current":
            path_output = self.path_output
        else:
            path_output = self.path_output

        if not os.path.exists(path_output):
            os.makedirs(path_output)
        self.path_output = standard_path(path_output)

        return self._python_run_bib_md_tex(output_md, output_tex, data_list_md_tex, original_bib_data)

    def _python_run_bib_md_tex(
        self,
        output_md: str,
        output_tex: str,
        data_list_md_tex: List[str],
        original_bib_data: Union[List[str], str, Library],
    ) -> Tuple[List[str], List[str]]:
        """Process BibTeX, Markdown and LaTeX content.

        Parameters
        ----------
        output_md : str
            Output Markdown filename
        output_tex : str
            Output LaTeX filename
        data_list_md_tex : List[str]
            List of content lines (Markdown or LaTeX)
        original_bib_data : Union[List[str], str, Library]
            BibTeX data in various formats

        Returns
        -------
        Tuple[List[str], List[str]]
            Tuple containing processed Markdown content and LaTeX content
        """
        # Copy figures if enabled
        if self.shutil_figures:
            figure_names = self.search_figure_names(data_list_md_tex)
            self.shutil_copy_figures(self.figure_folder_name, self.path_figures, figure_names, self.path_output)

        # Extract citation keys from content
        key_in_md_tex = self.search_cite_keys(data_list_md_tex, self.tex_md_flag)

        # Process bibliography
        full_bib_for_zotero, full_bib_for_abbr, full_bib_for_save = "", "", ""
        if key_in_md_tex:
            # Generate bib contents
            abbr_library, zotero_library, save_library = self._python_bib.parse_to_multi_standard_library(
                original_bib_data, key_in_md_tex
            )
            key_in_md_tex = sorted(list(abbr_library.entries_dict.keys()), key=key_in_md_tex.index)

            # Write bibliography files
            _path_output = os.path.join(self.path_output, self.bib_folder_name)
            full_bib_for_abbr, full_bib_for_zotero, full_bib_for_save = self._python_writer.write_multi_library_to_file(
                _path_output, abbr_library, zotero_library, save_library, key_in_md_tex
            )

        # Process content based on format
        if self.tex_md_flag == ".md":
            # Write original markdown content
            write_list(data_list_md_tex, output_md, "w", os.path.join(self.path_output, self.md_folder_name), False)

            # Generate processed content and write to given files
            data_list_md, data_list_tex = self._python_md.special_operate_for_md(
                self.path_output,
                data_list_md_tex,
                output_md,
                full_bib_for_abbr,
                full_bib_for_zotero,
                self.template_name,
                self.generate_html,
                self.generate_tex,
            )
        else:
            data_list_md, data_list_tex = [], data_list_md_tex

        # Generate LaTeX output if enabled
        if self.generate_tex:
            self._python_tex.generate_standard_tex_data_list(
                data_list_tex,
                output_tex,
                self.path_output,
                self.figure_folder_name,
                self.tex_folder_name,
                self.bib_folder_name,
                os.path.basename(full_bib_for_abbr),
                self.template_name,
            )

        # Cleanup original files if enabled
        if self.delete_original_md_in_output_folder:
            self._cleanup_file(os.path.join(self.path_output, self.md_folder_name, output_md))

        if self.delete_original_tex_in_output_folder:
            self._cleanup_file(os.path.join(self.path_output, self.tex_folder_name, output_tex))

        if self.delete_original_bib_in_output_folder:
            for file in [full_bib_for_abbr, full_bib_for_zotero, full_bib_for_save]:
                self._cleanup_file(file)

        return data_list_md, data_list_tex

    @staticmethod
    def search_figure_names(data_list: List[str], figure_postfixes: Optional[List[str]] = None) -> List[str]:
        """Search for figure filenames in content.

        Parameters
        ----------
        data_list : List[str]
            List of content lines to search
        figure_postfixes : Optional[List[str]], optional
            List of figure file extensions to look for, by default None

        Returns
        -------
        List[str]
            List of found figure filenames
        """
        if figure_postfixes is None:
            figure_postfixes = ["eps", "jpg", "png", "svg", "psd", "raw", "jpeg", "pdf"]

        regex = re.compile(rf'[\w\-]+\.(?:{"|".join(figure_postfixes)})', re.I)
        figure_names = []
        for line in data_list:
            figure_names.extend(regex.findall(line))
        return sorted(set(figure_names), key=figure_names.index)

    @staticmethod
    def shutil_copy_figures(fig_folder_name: str, path_fig: str, fig_names: List[str], path_output: str) -> None:
        """Copy figure files to output directory.

        Parameters
        ----------
        fig_folder_name : str
            Name of figures folder in output directory
        path_fig : str
            Source directory containing figures
        fig_names : List[str]
            List of figure filenames to copy
        path_output : str
            Output directory path

        Returns
        -------
        None
        """
        if not os.path.exists(path_fig):
            print(f"{path_fig} does not existed.")
            return None

        file_list = []
        for root, _, files in os.walk(path_fig, topdown=False):
            for name in files:
                if name in fig_names:
                    file_list.append(os.path.join(root, name))

        for file in file_list:
            path_output_file = os.path.join(path_output, fig_folder_name, os.path.basename(file))
            p = os.path.dirname(path_output_file)
            if not os.path.exists(p):
                os.makedirs(p)
            shutil.copy(file, path_output_file)
        return None

    @staticmethod
    def search_cite_keys(data_list: List[str], tex_md_flag: str = ".tex") -> List[str]:
        r"""Extract citation keys from content according to their places.

        Parameters
        ----------
        data_list : List[str]
            List of content lines to search
        tex_md_flag : str, optional
            Flag indicating content format (".tex" or ".md"), by default ".tex"

        Returns
        -------
        List[str]
            List of found citation keys

        Notes
        -----
        For LaTeX, searches for \\cite, \\citep, \\citet patterns
        For Markdown, searches for [@key], @key; and ;@key] patterns
        """
        cite_key_list = []
        if tex_md_flag == ".tex":
            regex_list = [re.compile(r"\\[a-z]*cite[tp]*{\s*([\w\-.,:/\s]*)\s*}")]
            cite_key_list.extend(regex_list[0].findall("".join(data_list)))
            cite_key_list = combine_content_in_list([re.split(",", c) for c in cite_key_list])
        elif tex_md_flag == ".md":
            regex_list = [
                re.compile(r"\[@([\w\-.:/]+)\]"),
                re.compile(r"@([\w\-.:/]+)\s*;"),
                re.compile(r";\s*@([\w\-.:/]*)\s*]"),
            ]
            cite_key_list = combine_content_in_list(
                [regex_list[i].findall("".join(data_list)) for i in range(len(regex_list))]
            )
        else:
            print(f"{tex_md_flag} must be `.tex` or `.md`.")

        cite_key_list = [c.strip() for c in cite_key_list if c.strip()]
        return sorted(set(cite_key_list), key=cite_key_list.index)

    def _cleanup_file(self, file_path: str) -> None:
        """Cleanup files and empty directories.

        Parameters
        ----------
        file_path : str
            Path to file to be removed

        Returns
        -------
        None
        """
        if os.path.exists(file_path):
            os.remove(file_path)
            dir_path = os.path.dirname(file_path)
            if dir_path != self.path_output:  # Don't remove the main output directory
                if len([f for f in os.listdir(dir_path) if f != ".DS_Store"]) == 0:
                    shutil.rmtree(dir_path)



from typing import List, Optional


class ParsingException(Exception):
    """Generic Exception for parsing errors."""

    def __copy__(self):
        # We do not copy or deepcopy ParsingExceptions
        # because they are used as immutables,
        # and because default memo fails.
        return self

    def __deepcopy__(self, memo):
        # We do not copy or deepcopy ParsingExceptions
        # because they are used as immutables,
        # and because default memo fails.
        return self


class BlockAbortedException(ParsingException):
    """Exception where a invalid bibtex file led to an aborted block."""

    def __init__(
        self,
        abort_reason: str,
        # Not provided if end of file is reached
        end_index: Optional[int] = None,
    ):
        self.abort_reason = abort_reason
        self.end_index = end_index


class ParserStateException(ParsingException):
    """Parser is in a self-inflicted invalid state."""

    def __init__(self, message: str):
        self.message = message


class RegexMismatchException(ParserStateException):
    """Raised when regex matches are inconsistent, implying a bug in the parser.

    For example, raised when first match ``@string{``
    is not followed by an overlapping match ``}``.
    """

    def __init__(self, first_match, expected_match, second_match):
        self.first_match = first_match
        self.expected_match = expected_match
        self.second_match = second_match
        super().__init__(
            f"Regex mismatch: {first_match} followed by {second_match},"
            f"but expected {expected_match}.\n"
            "This is an python-bibtexparser internal error. "
            "Please report this issue at our issue tracker."
        )


class PartialMiddlewareException(ParsingException):
    """Exception raised when a middleware could not be fully applied."""

    def __init__(self, reasons: List[str]):
        reasons_string = "\n\n=====\n\n".join(reasons)
        super().__init__(f"Middleware could not be fully applied: {reasons_string}")



import re
from typing import List, Tuple


class StandardizeCommentBlock(object):
    """Stanndardize comment block."""

    def __init__(self) -> None:
        pass

    def standardize(self, block: List[str]) -> Tuple[List[str], List[List[str]]]:
        implicit_comments = []

        regex_comment = re.compile(r"@comment{" + r"(.*)", re.DOTALL)
        if mch := regex_comment.match("".join(block)):
            a = mch.group(1).strip()
            if (ll := (a.count("{") + 1)) > (lr := a.count("}")):
                a += "}" * (ll - lr)
            elif ll < lr:
                a = "{" * (lr - ll) + a

            if sub_mch := re.match(r"(.*)" + "}" + r"(.*)(\n*)", a):
                block = []

                sub_a, sub_b, sub_c = sub_mch.groups()
                if sub_a.strip():
                    block = ["@comment{" + sub_a.replace("\n", " ").strip() + "}\n"]

                if sub_b.strip():
                    implicit_comments = [[sub_b + sub_c, __class__.__name__]]

            else:
                block = []
                implicit_comments = [["".join(block), __class__.__name__]]
        else:
            block = []
            implicit_comments = [["".join(block), __class__.__name__]]
        return block, implicit_comments



import copy
import math
import os
import re
from typing import Any, Dict, List, Union

from pyadvtools import (
    IterateCombineExtendDict,
    read_list,
    sort_int_str,
    standard_path,
    write_list,
)

from ..bib.bibtexparser import Block, Library
from ..main import PythonRunBib, PythonWriters
from ..tools.experiments_base import generate_readme


def format_entries_for_abbr_zotero_save(
    j_conf_abbr: str,
    path_output: str,
    original_data: Union[List[str], str, Library],
    combine_year_length: int = 1,
    default_year_list: List[str] = [],
    write_flag_bib: str = "w",
    check_bib_exist: bool = False,
    write_flag_readme: str = "w",
    check_md_exist: bool = False,
    options: Dict[str, Any] = {},
) -> None:
    """Format bibliography entries and organize them by year and type.

    Processes bibliography data and organizes it into separate files by entry type and year,
    generating both BibTeX files and README documentation.

    Parameters
    ----------
    j_conf_abbr : str
        Journal/conference abbreviation used for naming output files
    path_output : str
        Output directory path for processed files
    original_data : Union[List[str], str, Library]
        Input bibliography data in various formats (list of strings, file path, or Library object)
    combine_year_length : int, optional
        Number of years to combine in each output file, by default 1
    default_year_list : List[str], optional
        Specific years to process (if empty, processes all years), by default []
    write_flag_bib : str, optional
        Write mode for BibTeX files ("w" for write, "a" for append), by default "w"
    check_bib_exist : bool, optional
        Whether to check if BibTeX files exist before writing, by default False
    write_flag_readme : str, optional
        Write mode for README files ("w" for write, "a" for append), by default "w"
    check_md_exist : bool, optional
        Whether to check if README files exist before writing, by default False
    options : Dict[str, Any], optional
        Additional processing options, by default {}

    Returns
    -------
    None
    """
    path_output = standard_path(path_output)

    # Set up processing options
    _options = {}
    _options.update(options)
    _options["is_sort_entry_fields"] = True  # Force field sorting
    _options["is_sort_blocks"] = True  # Force block sorting
    _options["sort_entries_by_field_keys_reverse"] = False  # Sort in ascending order

    # Initialize helper classes
    _python_bib = PythonRunBib(_options)

    _options["empty_entry_cite_keys"] = True  # Allow empty citation keys
    _python_writer = PythonWriters(_options)

    # Organize entries by type, year, volume, number, and month
    entry_type_year_volume_number_month_entry_dict = _python_bib.parse_to_nested_entries_dict(original_data)

    # Process each entry type separately
    for entry_type in entry_type_year_volume_number_month_entry_dict:

        # Filter years if specified
        year_dict = entry_type_year_volume_number_month_entry_dict[entry_type]
        year_list = sort_int_str(list(year_dict.keys()))
        if default_year_list:
            year_list = [y for y in year_list if y in default_year_list]
        year_dict = {year: year_dict[year] for year in year_list}

        # Save bibliography files grouped by years
        path_write = os.path.join(path_output, entry_type.lower(), "bib")
        for i in range(math.ceil(len(year_list) / combine_year_length)):

            # Determine year range for this file
            start_year_index = i * combine_year_length
            end_year_index = min([(i + 1) * combine_year_length, len(year_list)])
            combine_year = year_list[start_year_index:end_year_index]

            # Create subset dictionary for these years
            new_year_dict = {year: year_dict[year] for year in combine_year}
            entries: List[Block] = IterateCombineExtendDict().dict_update(copy.deepcopy(new_year_dict))

            # Generate filename based on year range
            name = f"{j_conf_abbr}_{combine_year[0]}"
            if len(combine_year) > 1:
                name += f"_{combine_year[-1]}"
            name += ".bib"

            # Write the bibliography file
            _python_writer.write_to_file(entries, name, write_flag_bib, path_write, check_bib_exist)

        # Generate and save README documentation
        path_write = os.path.join(path_output, entry_type.lower())
        readme_md = generate_readme(j_conf_abbr, entry_type, year_dict)

        # Handle append mode for README
        if re.search("a", write_flag_readme):
            old_readme_md = [re.sub(r"[ ]+", "", line) for line in read_list("README.md", "r", path_write)]
            readme_md = readme_md[3:] if old_readme_md else readme_md
            readme_md = [line for line in readme_md if re.sub(r"[ ]+", "", line) not in old_readme_md]

        write_list(readme_md, "README.md", write_flag_readme, path_write, check_md_exist)


def generate_statistic_information(path_output: str) -> None:
    """Generate statistical information from bibliography files.

    Processes all BibTeX files in the directory tree and extracts key information
    (DOIs and URLs) into CSV files for analysis.

    Parameters
    ----------
    path_output : str
        Root directory containing BibTeX files to process

    Returns
    -------
    None
    """
    # Find all BibTeX files in the directory tree
    full_files = []
    for root, _, files in os.walk(path_output):
        full_files.extend([os.path.join(root, f) for f in files if f.endswith(".bib")])

    # Configure processing options
    _options = {
        "is_standardize_bib": False,  # Skip standardization, default is True
        "choose_abbr_zotero_save": "save",  # Use save format, default is "save"
        "function_common_again": False,  # Skip reprocessing, default is True
        "function_common_again_abbr": False,  # Skip abbreviation reprocessing, default is True
        "function_common_again_zotero": False,  # Skip Zotero reprocessing, default is True
        "function_common_again_save": False,  # Skip save format reprocessing, default is True
        "is_sort_entry_fields": False,  # Skip field sorting
        "is_sort_blocks": False,  # Skip block sorting
    }
    _python_bib = PythonRunBib(_options)

    # Process each BibTeX file
    for f in full_files:
        informations = []
        library = _python_bib.parse_to_single_standard_library(f)

        # Extract DOI or URL for each entry
        for entry in library.entries:
            flag = ""
            if not flag:
                flag = entry["doi"] if "doi" in entry else ""
            if not flag:
                flag = entry["url"] if "url" in entry else ""
            informations.append(flag + "\n")

        # Write information to CSV file
        csv_path = f.replace(".bib", ".csv").replace(f"{os.sep}bib{os.sep}", f"{os.sep}url{os.sep}")
        write_list(informations, csv_path, "w", None, False)

    return None



import re

from ...library import Library
from ...model import Block, Entry
from ..middleware import BlockMiddleware
from ..utils import generate_cite_key_prefix


class AddArchive(BlockMiddleware):
    """Add Field `archive`."""

    def __init__(
        self,
        full_abbr_article_dict: dict,
        full_abbr_inproceedings_dict: dict,
        full_names_in_json: str,
        abbr_names_in_json: str,
        allow_inplace_modification: bool = True
    ):
        super().__init__(allow_inplace_modification=allow_inplace_modification)

        self.full_abbr_article_dict = full_abbr_article_dict
        self.full_abbr_inproceedings_dict = full_abbr_inproceedings_dict
        self.full_names_in_json = full_names_in_json
        self.abbr_names_in_json = abbr_names_in_json

    # docstr-coverage: inherited
    def transform_entry(self, entry: Entry, library: Library) -> Block:
        entry["archive"] = generate_cite_key_prefix(
            entry,
            self.full_abbr_article_dict,
            self.full_abbr_inproceedings_dict,
            self.full_names_in_json,
            self.abbr_names_in_json
            )
        return entry


class AddJournalLongAbbr(BlockMiddleware):
    """Add long abbr for field `journal`."""

    def __init__(
        self,
        full_abbr_article_dict: dict,
        full_names_in_json: str,
        abbr_names_in_json: str,
        allow_inplace_modification: bool = True
    ):
        super().__init__(allow_inplace_modification=allow_inplace_modification)

        self.full_abbr_article_dict = full_abbr_article_dict
        self.full_names_in_json = full_names_in_json
        self.abbr_names_in_json = abbr_names_in_json

    # docstr-coverage: inherited
    def transform_entry(self, entry: Entry, library: Library) -> Block:
        return self.generate_journal_booktitle_long_abbr(entry)

    def generate_journal_booktitle_long_abbr(self, entry) -> Entry:
        # Only for journal
        if entry.entry_type.lower() == "article":
            full_abbr_dict = self.full_abbr_article_dict
            field_key = "journal"
        else:
            return entry

        # obtain new_dict
        abbr_dict_dict = {}
        for publisher in full_abbr_dict:
            abbr_dict_dict.update({abbr: full_abbr_dict[publisher][abbr] for abbr in full_abbr_dict[publisher]})

        field_content = entry[field_key] if field_key in entry else ""
        field_content = re.sub(r"\(.*\)", "", field_content).strip()

        if not field_content:
            return entry

        # match
        long_abbr_list = []
        for abbr in abbr_dict_dict:
            full_name_list = abbr_dict_dict[abbr].get(self.full_names_in_json, [])
            long_abbr_name_list = abbr_dict_dict[abbr].get(self.abbr_names_in_json, [])

            for full, long_abbr in zip(full_name_list, long_abbr_name_list):
                if re.match('{' + full + '}', '{' + field_content + '}', re.I):
                    long_abbr_list.append(long_abbr)

        # check
        long_abbr_list = list(set(long_abbr_list))
        if len(long_abbr_list) > 1:
            print(f"Multiple match: {long_abbr_list} for {field_content}.")
        elif len(long_abbr_list) == 1:
            entry["shortjournal"] = long_abbr_list[0]
        return entry



"""Initialization."""

__all__ = [
    "BasicInput",
    "PandocMdTo",
    "PythonRunBib",
    "PythonRunMd",
    "PythonRunTex",
    "PythonWriters",
]

from .basic_input import BasicInput
from .pandoc_md_to import PandocMdTo
from .python_run_bib import PythonRunBib
from .python_run_md import PythonRunMd
from .python_run_tex import PythonRunTex
from .python_writers import PythonWriters



FIELD_FORMAT_FLAG = [r'"', r"{", r"\w"]

DEFAULT_FIELDS_LIST = [
    "abstract",
    "accepted",
    "accessdate",
    "address",
    "affiliation",
    "affiliations",
    "annotation",
    "archive",
    "article-number",
    "articleno",
    "audio",
    "author",
    "author-email",
    "authors",
    "bibsource",
    "biburl",
    "book-group-author",
    "booktitle",
    "cited-references",
    "code",
    "copyright",
    "crossref",
    "data",
    "date",
    "doc-delivery-number",
    "doi",
    "edition",
    "editor",
    "editors",
    "eissn",
    "eprint",
    "extra",
    "fjournal",
    "funding-acknowledgement",
    "funding-text",
    "howpublished",
    "institution",
    "isbn",
    "isbn-13",
    "issn",
    "issue",
    "issue_date",
    "journal",
    "journal-iso",
    "journalabbr",
    "journalabbreviation",
    "key",
    "keywords",
    "keywords-plus",
    "language",
    "lccn",
    "location",
    "month",
    "note",
    "number",
    "number-of-cited-references",
    "numpages",
    "orcid-numbers",
    "organization",
    "pages",
    "pdf",
    "place",
    "pubdates",
    "publisher",
    "remark",
    "research-areas",
    "researcherid-numbers",
    "rights",
    "school",
    "series",
    "shortjournal"
    "shorttitle",
    "slide",
    "submitted",
    "summary",
    "supplementary",
    "times-cited",
    "timestamp",
    "title",
    "type",
    "unique-id",
    "url",
    "usage-count-last-180-days",
    "usage-count-since-2013",
    "video",
    "volume",
    "web-of-science-categories",
    "web-of-science-index",
    "xxeditor",
    "xxnote",
    "xxtitle",
    "xxyear",
    "year",
]



"""Initialization."""

__all__ = [
    "PyRunBibMdTex",
    "format_entries_for_abbr_zotero_save",
    "generate_statistic_information",

    "generate_standard_publisher_abbr_options_dict",

    "Searchkeywords",
    "generate_from_bibs_and_write",

    "PaperLinksGenerator",

    "compare_bibs_with_local",
    "compare_bibs_with_zotero",
    "replace_to_standard_cite_keys",

    "CheckDeleteFormatMoveSpideredBibs",
]

from .compare.compare_bibs import compare_bibs_with_local, compare_bibs_with_zotero
from .experiments_base import generate_standard_publisher_abbr_options_dict
from .format_save_bibs import format_entries_for_abbr_zotero_save, generate_statistic_information
from .generate.generate_from_bibs import generate_from_bibs_and_write
from .generate.generate_links import PaperLinksGenerator
from .py_run_bib_md_tex import PyRunBibMdTex
from .replace.replace import replace_to_standard_cite_keys
from .search.search_keywords import Searchkeywords
from .spider.process_spider_url_bib import CheckDeleteFormatMoveSpideredBibs



import abc
import logging
from copy import deepcopy
from typing import Collection, Union

from ..library import Library
from ..model import Block, Entry, ExplicitComment, ImplicitComment, Preamble, String

logger = logging.getLogger(__name__)


class Middleware(abc.ABC):
    """Implements a function to transform a block or library.

    Abstract Class. You should extend either BlockMiddleware
    or LibraryMiddleware
    """

    def __init__(
        self,
        allow_parallel_execution: bool = True,
        allow_inplace_modification: bool = True,
    ):
        """Create a new Middleware.

        :param allow_inplace_modification: See corresponding property.
        :param allow_parallel_execution: See corresponding property.
        """
        self._allow_inplace_modification = allow_inplace_modification
        self._allow_parallel_execution = allow_parallel_execution

    @property
    def allow_inplace_modification(self) -> bool:
        """If true, the middleware **may** modify the block in-place.

        I.e., if true, the output of `transform` may be the same instance
        as the input. If false, new instances must be returned.
        """
        return self._allow_inplace_modification

    @property
    def allow_parallel_execution(self) -> bool:
        """True indicates that the middleware is threadsafe."""
        return self._allow_parallel_execution

    @abc.abstractmethod
    def transform(self, library: Library) -> Library:
        """Apply transformation to a library. Main entrypoint of the middleware."""
        raise NotImplementedError("called abstract method")


class BlockMiddleware(Middleware, abc.ABC):
    """Transform a library on a per-block basis.

    The `BlockMiddleware` replaces a block with zero, one or more
    new (transformed) blocks.

    Changes may rely on the state of the overall library,
    but must not change the state of the library directly,
    except if `allow_inplace_modification` is true.
    """

    @classmethod
    def metadata_key(cls) -> str:
        """Tt is the identifier of the middleware.

        This key is used to identify the middleware in a blocks metadata.
        """
        return cls.__name__

    # docstr-coverage: inherited
    def transform(self, library: Library) -> Library:
        # TODO Multiprocessing (only for large library and if allow_multi..)
        blocks = []
        for b in library.blocks:
            transformed = self.transform_block(b, library)
            # Case 1: None. Skip it.
            if transformed is None:
                pass

            # Case 2: A single block. Add it to the list.
            elif isinstance(transformed, Block):
                blocks.append(transformed)

            # Case 3: A collection. Append all the elements.
            elif isinstance(transformed, Collection):
                # check that all the items are indeed blocks
                for item in transformed:
                    if not isinstance(item, Block):
                        raise TypeError(
                            f"Non-Block type found in transformed collection: {type(item)}"
                        )
                blocks.extend(transformed)

            # Case 4: Something else. Error.
            else:
                raise TypeError(f"Illegal output type from transform_block: {type(transformed)}")
        return Library(blocks=blocks)

    def transform_block(
        self, block: Block, library: Library
    ) -> Union[Block, Collection[Block], None]:
        """Transform a block.

        :param block: Block to transform.
        :param library: Library containing the block.
            Should typically not be modified during
            the transformation, but be considered as read-only.
            If the library is modified, make sure to set the `allow_multithreading`
            constructor argument to false
        :return: Transformed block. If the block should be removed, return None.
            If the block should be replaced by multiple blocks, return a collection
            of blocks. If the block should be replaced by a single block, return
            the single block. If the block should not be modified, return a copy of
            the original block.
            The returned block has to be a new instance, except if
            `self.allow_inplace_modification` is True (in which case the block
            may also return the original block).
        """
        block = block if self.allow_inplace_modification else deepcopy(block)
        if isinstance(block, Entry):
            return self.transform_entry(block, library)
        elif isinstance(block, String):
            return self.transform_string(block, library)
        elif isinstance(block, Preamble):
            return self.transform_preamble(block, library)
        elif isinstance(block, ExplicitComment):
            return self.transform_explicit_comment(block, library)
        elif isinstance(block, ImplicitComment):
            return self.transform_implicit_comment(block, library)

        # logger.warning(f"Unknown block type {type(block)}")
        return block

    def transform_entry(
        self, entry: Entry, library: Library
    ) -> Union[Block, Collection[Block], None]:
        """Transform an entry. Called by `transform_block` if the block is an entry.

        Note: This method modifies the passed entry. For a method
        respecting the `allow_inplace_modification` property,
        you should use `transform` or `transform_block` instead.
        """
        return entry

    def transform_string(
        self, string: String, library: Library
    ) -> Union[Block, Collection[Block], None]:
        """Transform a string. Called by `transform_block` if the block is a string.

        Note: This method modifies the passed string. For a method
        respecting the `allow_inplace_modification` property,
        you should use `transform` or `transform_block` instead.
        """
        return string

    def transform_preamble(
        self, preamble: Preamble, library: Library
    ) -> Union[Block, Collection[Block], None]:
        """Transform a preamble. Called by `transform_block` if the block is a preamble.

        Note: This method modifies the passed preamble. For a method
        respecting the `allow_inplace_modification` property,
        you should use `transform` or `transform_block` instead.
        """
        return preamble

    def transform_explicit_comment(
        self, explicit_comment: ExplicitComment, library: Library
    ) -> Union[Block, Collection[Block], None]:
        """Transform an explicit comment. Called by `transform_block` if the block is an explicit comment.

        Note: This method modifies the passed explicit comment. For a method
        respecting the `allow_inplace_modification` property,
        you should use `transform` or `transform_block` instead.
        """
        return explicit_comment

    def transform_implicit_comment(
        self, implicit_comment: ImplicitComment, library: Library
    ) -> Union[Block, Collection[Block], None]:
        """Transform an implicit comment. Called by `transform_block` if the block is an implicit comment.

        Note: This method modifies the passed implicit comment. For a method
        respecting the `allow_inplace_modification` property,
        you should use `transform` or `transform_block` instead.
        """
        return implicit_comment


class LibraryMiddleware(Middleware, abc.ABC):
    """Changes an overall library at once (not just on a per-block basis).

    Examples of library-wide changes are:
    - Re-Sorting the blocks in the library.
    - Transforming the library instance to a custom subclass of Library.

    Whatever can be done in a BlockMiddleware, should be done
    in a BlockMiddleware (and not in a LibraryMiddleware),
    for performance reasons (e.g. deleting blocks, ...).
    """

    def __init__(self, allow_inplace_modification: bool = True):
        # As library middleware is run per library (not per block individually),
        #   it cannot be parallelized.
        super().__init__(
            allow_inplace_modification=allow_inplace_modification,
            allow_parallel_execution=False,
        )

    def transform(self, library: Library) -> Library:
        """Transform a library.

        :param library: Library to transform.
        :return: Transformed library. If the library should not be modified,
            return a copy of the original library.
            The returned library has to be a new instance, except if
            `self.allow_inplace_modification` is True (in which case the library
            may also return the original library).
        """
        library = library if self.allow_inplace_modification else deepcopy(library)
        return library



import copy
import os
import re
from typing import Any, Dict, List, Optional, Tuple, Union

from pyadvtools import standard_path, transform_to_data_list

from ...bib.bibtexparser import Block, Library
from ...main import PythonRunBib, PythonWriters
from ..experiments_base import obtain_local_abbr_paths

ARXIV_BIORXIV = ["arxiv", "biorxiv", "ssrn"]


def obtain_local_abbr_paths_for_abbr(options: dict, path_spidered_bibs: str, path_spidering_bibs: str) -> List[str]:
    path_spidered_bibs = standard_path(path_spidered_bibs)
    path_spidering_bibs = standard_path(path_spidering_bibs)

    path_abbrs = []
    path_abbrs.extend(obtain_local_abbr_paths(os.path.join(path_spidered_bibs, "Journals"), options))
    path_abbrs.extend(obtain_local_abbr_paths(os.path.join(path_spidered_bibs, "Conferences"), options))
    path_abbrs.extend(obtain_local_abbr_paths(os.path.join(path_spidering_bibs, "spider_j"), options))
    path_abbrs.extend(obtain_local_abbr_paths(os.path.join(path_spidering_bibs, "spider_c"), options))

    if options.get("include_early_access", True):
        path_abbrs.extend(obtain_local_abbr_paths(os.path.join(path_spidering_bibs, "spider_j_e"), options))

    path_abbrs = [p for p in path_abbrs if os.path.basename(p).lower() not in ARXIV_BIORXIV]
    return path_abbrs


def compare_bibs_with_local(
    original_data: Union[List[str], str],
    path_spidered_bibs: str,
    path_spidering_bibs: str,
    path_output: str,
    options: Dict[str, Any],
) -> None:
    path_output = standard_path(path_output)

    # generate for original data
    _options = {}
    _options.update(options)
    _python_bib = PythonRunBib(_options)
    data_list = transform_to_data_list(original_data, ".bib")
    library = _python_bib.parse_to_single_standard_library(data_list)
    original_entry_keys = [entry.key for entry in library.entries]

    # generate dict for abbr key entry
    abbr_key_entries_dict, not_in_local_entries = generate_abbr_key_entry_dict(library, options)

    # compare with local bibs
    tuple_entries = _compare_with_local(abbr_key_entries_dict, path_spidered_bibs, path_spidering_bibs, options)
    searched_entries, not_searched_entries, duplicate_original_entries, duplicate_searched_entries = tuple_entries
    not_in_local_entries.extend(not_searched_entries)

    # write with sorting blocks according to original cite keys
    _options = {}
    _options["is_sort_entry_fields"] = True  # default is True
    _options["is_sort_blocks"] = True  # default is True
    _options["sort_entries_by_cite_keys"] = original_entry_keys
    _python_write = PythonWriters(_options)
    _python_write.write_to_file(searched_entries, "in_local_entries.bib", "w", path_output, False)
    _python_write.write_to_file(not_in_local_entries, "not_in_local_entries.bib", "w", path_output, False)

    # write without sorting blocks
    _options = {}
    _options["is_sort_entry_fields"] = True  # default is True
    _options["is_sort_blocks"] = False  # default is True
    _python_write = PythonWriters(_options)
    _python_write.write_to_file(duplicate_original_entries, "duplicate_original_entries.bib", "w", path_output, False)
    _python_write.write_to_file(duplicate_searched_entries, "duplicate_searched_entries.bib", "w", path_output, False)
    return None


def generate_abbr_key_entry_dict(library: Library, options: Dict[str, Any]):
    _options = {}
    _options["is_standardize_bib"] = True  # default is True
    _options["choose_abbr_zotero_save"] = "save"  # default is "save"
    _options["function_common_again"] = True  # default is True
    _options["generate_entry_cite_keys"] = True  # default is False
    _options.update(options)
    _python_bib = PythonRunBib(_options)

    abbr_key_entries_dict, not_in_local_entries = {}, []
    for entry in library.entries:
        flag = False

        if ("title" in entry) and (entry["title"].strip()) and ("year" in entry) and (entry["year"].strip()):
            temp_library = _python_bib.parse_to_single_standard_library(copy.deepcopy(Library([entry])))
            if len(entries := temp_library.entries) == 1:

                # article and inproceedings
                temps = entries[0].key.split("_")
                if (len(temps) == 3) and temps[0].lower() in ["j", "c"]:
                    abbr_key_entries_dict.setdefault(temps[1], {}).update({entry.key: entry})
                    flag = True

                # misc (arXiv, bioRxiv, and ssrn)
                elif (len(temps) == 2) and temps[0].lower() in ARXIV_BIORXIV:
                    abbr_key_entries_dict.setdefault(temps[0], {}).update({entry.key: entry})
                    flag = True

        if not flag:
            not_in_local_entries.append(entry)
    return abbr_key_entries_dict, not_in_local_entries


def _compare_with_local(
    abbr_key_entries_dict: Dict[str, Dict[str, Block]],
    local_path_spidered_bibs: str,
    local_path_spidering_bibs: str,
    options: Dict[str, Any],
) -> Tuple[List[Block], List[Block], List[Block], List[Block]]:
    # compare with local bibs
    searched_entries, not_searched_entries, duplicate_original_entries, duplicate_searched_entries = [], [], [], []
    for abbr, old_key_entries_dict in abbr_key_entries_dict.items():

        options_ = {}
        options_.update(options)
        if abbr.lower() not in ARXIV_BIORXIV:
            options_["include_abbr_list"] = [abbr]
        path_abbrs = obtain_local_abbr_paths_for_abbr(options_, local_path_spidered_bibs, local_path_spidering_bibs)

        new_key_entries_dict = {}
        for path_abbr in path_abbrs:
            if len(data_list := transform_to_data_list(path_abbr, ".bib")) == 0:
                continue

            print("*" * 9 + f" Compare in {f'{os.sep}'.join(path_abbr.split(os.sep)[-3:])} for {abbr} " + "*" * 9)

            _options = {}
            _options["is_standardize_bib"] = False  # default is True
            _options["choose_abbr_zotero_save"] = "save"  # default is "save"
            _options["function_common_again"] = False  # default is True
            _options["function_common_again_abbr"] = False  # default is True
            _options["function_common_again_zotero"] = False  # default is True
            _options["function_common_again_save"] = False  # default is True
            _options["generate_entry_cite_keys"] = False  # default is False
            _options.update(options)
            _library = PythonRunBib(_options).parse_to_single_standard_library(data_list)
            for key, entry in old_key_entries_dict.items():
                for _entry in _library.entries:
                    if check_equal_for_entry(entry, _entry, ["title"], abbr):
                        new_key_entries_dict.setdefault(key, []).append(copy.deepcopy(_entry))

        print()

        for key, entry in old_key_entries_dict.items():
            entries = new_key_entries_dict.get(key, [])
            if (length := len(entries)) == 1:
                entries[0].key = key
                searched_entries.extend(entries)
            elif length == 0:
                not_searched_entries.append(entry)
            else:
                for i, _entry in enumerate(entries):
                    _entry.key = key + "-a" * i
                duplicate_original_entries.append(entry)
                duplicate_searched_entries.extend(entries)

    return searched_entries, not_searched_entries, duplicate_original_entries, duplicate_searched_entries


def check_equal_for_entry(original_entry, new_entry, compare_field_list: List[str], abbr: Optional[str] = None):
    a_list, b_list = [original_entry.entry_type.lower()], [new_entry.entry_type.lower()]
    if (abbr is not None) and (abbr.lower() in ARXIV_BIORXIV):
        a_list, b_list = [], []

    regex_title = re.compile(r"\\href{(.*)}{(.*)}")
    for field in compare_field_list:
        x = original_entry[field].lower().strip() if field in original_entry else ""
        y = new_entry[field].lower().strip() if field in new_entry else ""

        if field == "title":
            if mch := regex_title.search(x):
                x = mch.group(2)
            if mch := regex_title.search(y):
                y = mch.group(2)

        a_list.append(x)
        b_list.append(y)

    a_list = [re.sub(r"\W", "", a) for a in a_list]
    b_list = [re.sub(r"\W", "", b) for b in b_list]
    if "_".join(a_list) == "_".join(b_list):
        return True
    return False


def compare_bibs_with_zotero(
    zotero_bib: Union[List[str], str],
    download_bib: Union[List[str], str],
    path_output: str,
    options: Dict[str, Any],
) -> None:
    path_output = standard_path(path_output)

    # for zotero bib
    _options = {}
    _options.update(options)
    _options["generate_entry_cite_keys"] = False  # default is False
    _python_bib = PythonRunBib(_options)
    data_list = transform_to_data_list(zotero_bib, ".bib")
    zotero_library = _python_bib.parse_to_single_standard_library(data_list)

    # for download bib
    _options = {}
    _options.update(options)
    _options["generate_entry_cite_keys"] = True  # default is False
    _python_bib = PythonRunBib(_options)
    data_list = transform_to_data_list(download_bib, ".bib")
    download_library = _python_bib.parse_to_single_standard_library(data_list)

    # compare download bib and zotero bib
    only_in_download_entries, in_download_and_zotero_entries = [], []
    for download_entry in download_library.entries:
        flag = False
        for zotero_entry in zotero_library.entries:
            if check_equal_for_entry(zotero_entry, download_entry, ["title"], None):
                in_download_and_zotero_entries.append(download_entry)
                flag = True
                break

        if not flag:
            only_in_download_entries.append(download_entry)

    # write
    _options = {}
    _options.update(options)
    _python_write = PythonWriters(_options)
    _python_write.write_to_file(only_in_download_entries, "only_in_download.bib", "w", path_output, False)
    _python_write.write_to_file(in_download_and_zotero_entries, "in_download_and_zotero.bib", "w", path_output, False)
    return None



from ...library import Library
from ...model import Block, Entry
from ..middleware import BlockMiddleware


class NormalizeEntryTypes(BlockMiddleware):
    """Normalize Entry types."""

    def __init__(self, allow_inplace_modification: bool = True):
        super().__init__(allow_inplace_modification=allow_inplace_modification, allow_parallel_execution=True)

    # docstr-coverage: inherited
    def transform_entry(self, entry: Entry, library: Library) -> Block:
        entry.entry_type = entry.entry_type.lower()
        return entry



from copy import deepcopy
from typing import Any, Dict, List, Optional, Union

from ..bibtexparser import (
    BibtexFormat,
    Block,
    DuplicateBlockKeyBlock,
    Entry,
    ExplicitComment,
    Field,
    ImplicitComment,
    Library,
    MiddlewaresLibraryToStr,
    ParsingFailedBlock,
    Preamble,
    String,
)
from .convert_library_to_library import ConvertLibrayToLibrary

VAL_SEP = " = "
PARSING_FAILED_COMMENT = "% WARNING Parsing failed for the following {n} lines."


class ConvertLibrayToStr(object):
    """Convert library to str.

    Args:
        options (Dict[str, Any]): Options. Default is {}.

    Attributes:
        is_standardize_library (bool): Is standardize library. Default is False.
        empty_entry_cite_keys (bool): Empty entry cite keys. Default is False.
        add_index_to_entries (bool): Add index to entries. Default is False.
    """

    def __init__(self, options: Dict[str, Any]):

        self.is_standardize_library = options.get("is_standardize_library", False)
        self.empty_entry_cite_keys = options.get("empty_entry_cite_keys", False)
        self.add_index_to_entries = options.get("add_index_to_entries", False)
        self.entries_necessary = options.get("entries_necessary", True)

        self.options = options

    def generate_str(
        self, library: Union[Library, List[Block]], bibtex_format: Optional[BibtexFormat] = None
    ) -> List[str]:
        """Serialize a BibTeX database.

        :param library: BibTeX database to serialize.
        :param bibtex_format: Customized BibTeX format to use (optional).
        """
        # --------- --------- --------- #
        if not isinstance(library, Library):
            library = Library(library)

        # standardizer
        if self.is_standardize_library:
            library = ConvertLibrayToLibrary(self.options).generate_single_library(library)

        # --------- --------- --------- #
        library = MiddlewaresLibraryToStr(self.options).functions(library)

        # --------- --------- --------- #
        if bibtex_format is None:
            bibtex_format = BibtexFormat()

        if bibtex_format.value_column == "auto":
            auto_val: int = self._calculate_auto_value_align(library)
            # Copy the format instance to avoid modifying the original
            # (which would be bad if the format is used for multiple libraries)
            bibtex_format = deepcopy(bibtex_format)
            bibtex_format.value_column = auto_val

        # --------- --------- --------- #
        if self.entries_necessary:
            if not library.entries:
                return []

        data_list = []
        j = 0
        for i, block in enumerate(library.blocks):
            if self.add_index_to_entries and isinstance(block, Entry):
                data_list.append(f"% {j + 1}\n")
                j += 1

            # Get str representation of block
            pieces = self._treat_block(bibtex_format, block)
            data_list.extend(pieces)

            # Separate Blocks
            if i < len(library.blocks) - 1:
                data_list.append(bibtex_format.block_separator)
        return data_list

    @staticmethod
    def _calculate_auto_value_align(library: Library) -> int:
        max_key_len = 0
        for entry in library.entries:
            for key in entry.fields_dict:
                max_key_len = max(max_key_len, len(key))
        return max_key_len

    def _treat_block(self, bibtex_format, block) -> List[str]:
        if isinstance(block, Entry):
            pieces = self._treat_entry(block, bibtex_format)
        elif isinstance(block, String):
            pieces = self._treat_string(block, bibtex_format)
        elif isinstance(block, Preamble):
            pieces = self._treat_preamble(block, bibtex_format)
        elif isinstance(block, ExplicitComment):
            pieces = self._treat_expl_comment(block, bibtex_format)
        elif isinstance(block, ImplicitComment):
            pieces = self._treat_impl_comment(block, bibtex_format)
        elif isinstance(block, ParsingFailedBlock):
            pieces = self._treat_failed_block(block, bibtex_format)
        else:
            raise ValueError(f"Unknown block type: {type(block)} in {__file__}")
        return pieces

    # entry
    def _treat_entry(self, block: Entry, bibtex_format: BibtexFormat) -> List[str]:
        if self.empty_entry_cite_keys:
            result = ["@" + block.entry_type + "{" + " " + ",\n"]
        else:
            result = ["@" + block.entry_type + "{" + block.key + ",\n"]
        field: Field
        for i, field in enumerate(block.fields):
            res = []
            res.append(bibtex_format.indent)
            res.append(field.key)
            res.append(self._val_intent_string(bibtex_format, field.key))
            res.append(VAL_SEP)
            res.append("{")  # add by me
            res.append(field.value)
            res.append("}")  # add by me
            if bibtex_format.trailing_comma or i < len(block.fields) - 1:
                res.append(",")
            res.append("\n")
            result.append("".join(res))

        result.append("}\n")
        return result

    @staticmethod
    def _val_intent_string(bibtex_format: BibtexFormat, key: str) -> str:
        """Calculate the spaces which have to be added after the ` = `."""
        if isinstance(bibtex_format.value_column, int):
            length = bibtex_format.value_column - len(key)
            return "" if length <= 0 else " " * length
        else:
            return ""

    # string
    @staticmethod
    def _treat_string(block: String, bibtex_format: BibtexFormat) -> List[str]:
        result = [
            "@string{",
            block.key,
            VAL_SEP,
            "{",
            block.value,
            "}",
            "}\n",
        ]
        return ["".join(result)]

    # preamble
    @staticmethod
    def _treat_preamble(block: Preamble, bibtex_format: BibtexFormat) -> List[str]:
        result = ["@preamble{" + f' "{block.value.rstrip()} "' + " }", "\n"]
        return ["".join(result)]

    # implicit comment
    @staticmethod
    def _treat_impl_comment(block: ImplicitComment, bibtex_format: BibtexFormat) -> List[str]:
        # Note: No explicit escaping is done here - that should be done in middleware
        result = [block.comment.rstrip(), "\n"]
        return ["".join(result)]

    # explicit comment
    @staticmethod
    def _treat_expl_comment(block: ExplicitComment, bibtex_format: BibtexFormat) -> List[str]:
        result = ["@comment{", block.comment.rstrip(), "}\n"]
        return ["".join(result)]

    # failed block
    @staticmethod
    def _treat_failed_block(block: ParsingFailedBlock, bibtex_format: BibtexFormat) -> List[str]:
        if isinstance(block.raw, str):
            lines = len(block.raw.splitlines())
            parsing_failed_comment = PARSING_FAILED_COMMENT.format(n=lines)
            return [parsing_failed_comment.rstrip(), "\n", block.raw.rstrip(), "\n"]
        else:
            if isinstance(block, DuplicateBlockKeyBlock):
                print(f"Duplicate key block: previous block key is `{block.key}`.")
            else:
                print(f"The raw of the field block: {block} is None.")
            return []



import copy
import os
import re
import time

from pyadvtools import (
    IterateCombineExtendDict,
    iterate_obtain_full_file_names,
    read_list,
    standard_path,
    write_list,
)

from ...bib.bibtexparser.library import Library
from ...main import PythonRunBib, PythonWriters
from ..experiments_base import generate_readme
from ..format_save_bibs import format_entries_for_abbr_zotero_save, generate_statistic_information

EXCLUDE_ABBR_LIST = ["arxiv", "biorxiv", "ssrn"]


class ProcessSpiderBib(object):
    """Process spider bib.

    Args:
        path_abbr: The path of the abbreviation folder.
        abbr_standard: The standard abbreviation.

    Attributes:
        path_abbr: The path of the abbreviation folder.
        abbr_standard: The standard abbreviation.
    """

    def __init__(self, path_abbr: str, abbr_standard: str) -> None:
        self.path_abbr = os.path.expandvars(os.path.expanduser(path_abbr))
        self.abbr_standard = abbr_standard

        self._options = {
            "is_standardize_bib": True,  # default is True
            "substitute_old_list": [
                r"(<[a-zA-Z\-]+\s*/*\s*>)",
                r'(</[a-zA-Z\-]+>)',
                r'(<[a-zA-Z\-]+ [^\s]+="[^>]+?"\s*/*\s*>)',
                r"([ ]+)",
                r";[; ]*;",
                r",[, ]*,"
            ],
            "substitute_new_list": ["", "", "", " ", ";", ","],
            "is_sort_entry_fields": True,  # default is False
            "is_sort_blocks": True,  # default is False
            "sort_entries_by_field_keys_reverse": False,  # default is True
            "empty_entry_cite_keys": True,
        }

        self._python_bib = PythonRunBib(self._options)

    def format_spider_bib(self, write_bib: bool = False) -> None:
        """Format spider bib."""
        file_list = iterate_obtain_full_file_names(self.path_abbr, ".bib", False)

        if write_bib:
            if os.path.exists(readme := os.path.join(self.path_abbr, "README.md")):
                os.remove(readme)

        _options = {}
        _options.update(self._options)
        _python_writer = PythonWriters(_options)

        for f in file_list:
            print("*" * 5 + f" Format {os.path.basename(f)} " + "*" * 5)

            data_list = read_list(f, "r")

            # standardize
            entry_type_year_volume_number_month_entry_dict = self._python_bib.parse_to_nested_entries_dict(data_list)
            if not write_bib:
                continue

            # just for the necessary part
            old_readme_md = [re.sub(r"[ ]+", "", line) for line in read_list("README.md", "r", self.path_abbr)]
            new_readme_md = []
            new_entry_list = []

            for entry_type in entry_type_year_volume_number_month_entry_dict:
                new_dict = entry_type_year_volume_number_month_entry_dict.get(entry_type.lower(), {})

                # for README.md
                readme_md = generate_readme(self.abbr_standard, entry_type.lower(), new_dict)
                readme_md = readme_md[3:] if (old_readme_md or new_readme_md) else readme_md
                readme_md = [line for line in readme_md if re.sub(r"[ ]+", "", line) not in old_readme_md]
                new_readme_md.extend(readme_md)

                # for bib
                entry_list = IterateCombineExtendDict().dict_update(copy.deepcopy(new_dict))
                new_entry_list.extend(entry_list)

            write_list(new_readme_md, "README.md", "a", self.path_abbr, False)
            _python_writer.write_to_file(new_entry_list, f, "w", None, False)
        return None

    def check_spider_bib(self, delete_duplicate_in_bibs: bool = False) -> None:
        """Check bib."""
        bibs_name = iterate_obtain_full_file_names(self.path_abbr, ".bib", False)
        bibs_name = [[f, os.path.basename(f).split(".")[0].strip()] for f in bibs_name]

        urls_name = iterate_obtain_full_file_names(self.path_abbr, ".csv", False)
        urls_name = [[f, os.path.basename(f).split(".")[0].strip()] for f in urls_name]

        url_base_names = [name[-1] for name in urls_name]

        _options = {}
        _options.update(self._options)
        _python_writer = PythonWriters(_options)

        for name in bibs_name:
            bib_base_name = name[-1]
            if bib_base_name not in url_base_names:
                print(f"{bib_base_name}.csv not in the folder `url`.")
                continue

            full_bib, full_url = name[0], urls_name[url_base_names.index(bib_base_name)][0]

            print("*" * 5 + f" Check {os.path.basename(full_bib)} and {os.path.basename(full_url)} " + "*" * 5)
            bib_list = read_list(full_bib, "r")

            # Check duplicated blocks in bib file
            library = self._python_bib.parse_to_single_standard_library(bib_list)

            url_bib_dict = {}
            for entry in library.entries:
                doi = entry["doi"] if "doi" in entry else ""
                url_ = entry["url"] if "url" in entry else ""
                url = doi if doi else url_
                url_bib_dict.setdefault(url, []).append(entry)

            duplicate_url, new_entries = [], []
            for url in url_bib_dict:
                if len(url_bib_dict[url]) > 1:
                    duplicate_url.append(url)
                if delete_duplicate_in_bibs:
                    new_entries.append(url_bib_dict[url][0])

            # Delete duplicated blocks in bib file
            if duplicate_url:
                print(f"Duplicates in {full_bib}: {duplicate_url}\n")
            if duplicate_url and delete_duplicate_in_bibs:
                _python_writer.write_to_file(new_entries, full_bib, "w", None, False)
        return None

    def move_spider_bib(self, path_shutil: str) -> None:
        if self.abbr_standard.lower() in EXCLUDE_ABBR_LIST:
            return None

        # Move
        print("*" * 5 + f" Start moving {self.abbr_standard} ... " + "*" * 5)
        path_move = os.path.join(path_shutil, self.abbr_standard)
        entry_type_entry_dict = {}
        library = PythonRunBib({}).parse_to_single_standard_library(self.path_abbr)
        for entry in library.entries:
            entry_type_entry_dict.setdefault(entry.entry_type, []).append(entry)
        for entry_type in entry_type_entry_dict:
            format_entries_for_abbr_zotero_save(
                self.abbr_standard,
                path_move,
                Library(entry_type_entry_dict[entry_type]),
                combine_year_length=1,
                default_year_list=self._default_year_list(entry_type),
                write_flag_bib="a",
                check_bib_exist=False,
                write_flag_readme="a",
                check_md_exist=False,
                options=self._options,
            )
        generate_statistic_information(path_move)
        print("*" * 5 + " Successfully moving ... " + "*" * 5)

        # Delete
        _options = {}
        _options.update(self._options)
        _python_writer = PythonWriters(_options)
        print("*" * 5 + f" Start deleting {self.abbr_standard} ... " + "*" * 5)
        bibs = iterate_obtain_full_file_names(self.path_abbr, ".bib")
        for bib in bibs:
            new_entries = []
            library = self._python_bib.parse_to_single_standard_library(read_list(bib, "r"))
            for entry in library.entries:
                year = entry["year"] if "year" in entry else ""
                if year not in self._default_year_list(entry.entry_type):
                    new_entries.append(entry)
            _python_writer.write_to_file(new_entries, bib, "w", None, False, True, True, False, True)
        print("*" * 5 + " Successfully deleting ... " + "*" * 5)

    @staticmethod
    def _default_year_list(entry_type) -> list:
        year = int(time.strftime("%Y", time.localtime()))
        month = int(time.strftime("%m", time.localtime()))
        m = 0 if month <= 3 else 1
        if entry_type == "article":
            default_year_list = [str(i) for i in range(1800, year + m - 1)]
        elif entry_type == "inproceedings":
            default_year_list = [str(i) for i in range(1800, year + 2)]
        else:
            default_year_list = [str(i) for i in range(1800, year + m - 1)]
        return default_year_list

    def simplify_early_access(self):
        # for IEEE Early Access
        path_ieee_early_access = self.path_abbr
        path_ieee = path_ieee_early_access.replace("spider_j_e", "spider_j")

        _options = {}
        _options.update(self._options)
        _python_writer = PythonWriters(_options)

        print(f"***** Simplify {self.abbr_standard} *****")
        path_url_ieee_early_access = os.path.join(path_ieee_early_access, "url")
        path_bib_ieee_early_access = os.path.join(path_ieee_early_access, "bib")
        path_url_ieee = os.path.join(path_ieee, "url")

        # for txt urls
        data_list = read_list(f"{self.abbr_standard}_0.txt", "r", path_url_ieee_early_access)
        for name in [f for f in os.listdir(path_url_ieee) if f.endswith(".txt")]:
            temp_data_list = read_list(name, "r", path_url_ieee)
            data_list = list(set(data_list).difference(set(temp_data_list)))
        write_list(sorted(data_list), f"{self.abbr_standard}_0.txt", "w", path_url_ieee_early_access, False)

        # for csv urls
        data_list_csv = read_list(f"{self.abbr_standard}_0.csv", "r", path_url_ieee_early_access)
        data_list_txt = read_list(f"{self.abbr_standard}_0.txt", "r", path_url_ieee_early_access)
        data_list = list(set(data_list_csv).intersection(set(data_list_txt)))
        write_list(sorted(data_list), f"{self.abbr_standard}_0.csv", "w", path_url_ieee_early_access, False)

        # for bibs
        data_list_bib = read_list(f"{self.abbr_standard}_0.bib", "r", path_bib_ieee_early_access)
        data_list_url = read_list(f"{self.abbr_standard}_0.txt", "r", path_url_ieee_early_access)

        entries = []
        library = self._python_bib.parse_to_single_standard_library(data_list_bib)
        for url in data_list_url:
            for entry in library.entries:
                if standard_path(url) == standard_path(entry["url"]):
                    entries.append(entry)
                    break

        _python_writer.write_to_file(
            entries, f"{self.abbr_standard}_0.bib", "w", path_bib_ieee_early_access, False, True, True, True
        )



import json
import os
from typing import Any, Dict

from pyadvtools import read_list, standard_path


class BasicInput(object):
    """Basic input.

    Args:
        options (Dict[str, Any]): Options.

    Attributes:
        path_bibs (str): Path bibs.
        path_figures (str): Path figures.
        path_templates (str): Path templates.

        full_abbr_article_dict (Dict[str, str]): Full abbr article dict.
        full_abbr_inproceedings_dict (Dict[str, str]): Full abbr inproceedings dict.
        full_names_in_json (str): Full names in json.
        abbr_names_in_json (str): Abbr names in json.

        full_csl_style_pandoc (str): Full path to csl style for pandoc.
        full_tex_article_template_pandoc (str): Full path to tex article template for pandoc.
        article_template_tex (List[str]): Article template for LaTex.

        article_template_header_tex (List[str]): Article template header for LaTex.
        article_template_tail_tex (List[str]): Article template tail for LaTex.
        beamer_template_header_tex (List[str]): Beamer template header for LaTex.
        beamer_template_tail_tex (List[str]): Beamer template tail for LaTex.
        tex_math_commands_tex (List[str]): Tex math commands for LaTex.
        tex_usepackages_tex (List[str]): Tex usepackages for LaTex.
        handly_preamble (bool): Handly preamble.

        options (Dict[str, Any]): Options.
    """

    def __init__(self, options: Dict[str, Any]) -> None:
        # The paths of Figures, Bibs, and Templates
        self.path_bibs: str = standard_path(options.get("path_bibs", ""))
        self.path_figures: str = standard_path(options.get("path_figures", ""))
        self.path_templates: str = standard_path(options.get("path_templates", ""))

        # Update
        path_config = standard_path(options.get("path_config", ""))
        if len(self.path_bibs) == 0:
            for folder in [
                "bib", "bibs", "Bib", "Bibs", "BIB", "BIBS",
                "reference", "references", "Reference", "References", "REFERENCE", "REFERENCES"
            ]:
                if os.path.exists(p := os.path.join(path_config, folder)):
                    self.path_bibs = p
                    break

        if len(self.path_figures) == 0:
            for folder in [
                "figure", "figures", "Figure", "Figures", "FIGURE", "FIGURES",
                "fig", "figs", "Fig", "Figs", "FIG", "FIGS",
            ]:
                if os.path.exists(p := os.path.join(path_config, folder)):
                    self.path_figures = p
                    break

        if len(self.path_templates) == 0:
            for folder in ["template", "templates", "Template", "Templates", "TEMPLATE", "TEMPLATES"]:
                if os.path.exists(p := os.path.join(path_config, folder)):
                    self.path_templates = p
                    break

        # bib/core
        self._initialize_middlewares(options)

        # main
        self._initialize_pandoc_md_to(options)
        self._initialize_python_run_tex(options)

        self.options = options

    # bib/core
    def _initialize_middlewares(self, options: Dict[str, Any]) -> None:
        full_json_c = os.path.join(self.path_templates, "AbbrFull", "conferences.json")
        full_json_j = os.path.join(self.path_templates, "AbbrFull", "journals.json")
        if os.path.isfile(full_json_c):
            with open(full_json_c, "r") as f:
                try:
                    json_dict = json.loads(f.read())
                except Exception as e:
                    print(e)
                    json_dict = {}
                full_abbr_inproceedings_dict = {p: json_dict[p].get("conferences", {}) for p in json_dict}
        else:
            full_abbr_inproceedings_dict = {}

        if os.path.isfile(full_json_j):
            with open(full_json_j, "r") as f:
                try:
                    json_dict = json.loads(f.read())
                except Exception as e:
                    print(e)
                    json_dict = {}
                full_abbr_article_dict = {p: json_dict[p].get("journals", {}) for p in json_dict}
        else:
            full_abbr_article_dict = {}

        inproceedings_dict = options.get("full_abbr_inproceedings_dict", {})
        if len(inproceedings_dict) == 0:
            inproceedings_dict = full_abbr_inproceedings_dict
        self.full_abbr_inproceedings_dict = full_abbr_inproceedings_dict

        article_dict = options.get("full_abbr_article_dict", {})
        if len(article_dict) == 0:
            article_dict = full_abbr_article_dict
        self.full_abbr_article_dict = full_abbr_article_dict

        full_names_in_json = options.get("full_names_in_json", "names_full")
        if len(full_names_in_json) == 0:
            full_names_in_json = "names_full"
        self.full_names_in_json = full_names_in_json

        abbr_names_in_json = options.get("abbr_names_in_json", "names_abbr")
        if len(abbr_names_in_json) == 0:
            abbr_names_in_json = "names_abbr"
        self.abbr_names_in_json = abbr_names_in_json

        options["full_abbr_article_dict"] = self.full_abbr_article_dict
        options["full_abbr_inproceedings_dict"] = self.full_abbr_inproceedings_dict

        options["full_names_in_json"] = self.full_names_in_json
        options["abbr_names_in_json"] = self.abbr_names_in_json

    # main
    def _initialize_pandoc_md_to(self, options: Dict[str, Any]) -> None:
        csl_name = options.get("csl_name", "apa-no-ampersand")
        if len(csl_name) == 0:
            csl_name = "apa-no-ampersand"

        full_csl_style_pandoc = os.path.join(self.path_templates, "CSL", f"{csl_name}.csl")
        if (p := options.get("full_csl")) is not None:
            full_csl_style_pandoc = p
        self.full_csl_style_pandoc = full_csl_style_pandoc

        full_tex_article_template_pandoc = os.path.join(self.path_templates, "TEX", "eisvogel.tex")
        if (p := options.get("full_eisvogel")) is not None:
            full_tex_article_template_pandoc = p
        self.full_tex_article_template_pandoc = full_tex_article_template_pandoc

        self.article_template_tex = self._try_read_list(options, "TEX", "Article.tex", "full_article")

    def _initialize_python_run_tex(self, options: Dict[str, Any]) -> None:
        self.article_template_header_tex = self._try_read_list(options, "TEX", "Article_Header.tex", "full_article_header")
        self.article_template_tail_tex = self._try_read_list(options, "TEX", "Article_Tail.tex", "full_article_tail")
        self.beamer_template_header_tex = self._try_read_list(options, "TEX", "Beamer_Header.tex", "full_beamer_header")
        self.beamer_template_tail_tex = self._try_read_list(options, "TEX", "Beamer_Tail.tex", "full_beamer_tail")
        self.math_commands_tex = self._try_read_list(options, "TEX", "math_commands.tex", "full_math_commands")
        self.usepackages_tex = self._try_read_list(options, "TEX", "Style.tex", "full_usepackages_tex")

        # handly preamble
        self.handly_preamble = options.get("handly_preamble", False)
        if self.handly_preamble:
            self.article_template_header_tex, self.article_template_tail_tex = [], []
            self.beamer_template_header_tex, self.beamer_template_tail_tex = [], []
            self.math_commands_tex, self.usepackages_tex = [], []

    def _try_read_list(self, options: Dict[str, Any], folder_name: str, file_name: str, key: str):
        path_file = os.path.join(self.path_templates, folder_name, file_name)
        if (p := options.get(key)) is None:
            return []
        else:
            path_file = p

        try:
            data_list = read_list(path_file)
        except Exception as e:
            print(e)
            data_list = []
        return data_list



import os
import re

from pyadvtools import (
    combine_content_in_list,
    read_list,
    write_list,
)

html_head = """<!DOCTYPE html>
<html>
<head>
  <meta charset="utf-8">
  <title>{}</title>
"""

html_style = """  <style>
    html {font-size: 22px;}
    body {margin: 0 auto; max-width: 76em;}
    #copyID {font-size: 18px;}
  </style>
  <script>
    function copy(element) {
      if (element.type == "button"){
      element.type="text";
      }
      element.style.color="black";
      element.style.backgroundColor="#C7EDCC";
      element.select();
      element.setSelectionRange(0, 99999);
      navigator.clipboard.writeText(element.value);
      window.getSelection().removeAllRanges();
      element.type="button";
    }
  </script>
</head>
<body>
"""

html_tail = """
</body>
</html>
"""

textarea_header = '<textarea id="copyID" onclick="copy(this)" rows="16" cols="145">\n'
textarea_tail = "\n</textarea>"


def operate_on_generate_html(html_name: str) -> None:
    if not (data_list := read_list(html_name, "r", None)):
        return None

    head_list = [html_head.format(os.path.basename(html_name).split('.')[0].strip()), html_style, "\n"]
    tail_list = [html_tail]

    content = "".join(data_list)
    content = content.replace("<pre><code>", textarea_header).replace("</code></pre>", textarea_tail)
    for i in re.findall(r"<li>(.*?)<details>", content, re.DOTALL):
        content = content.replace(rf"<li>{i}<details>", f"<li><details>\n<summary>\n{i.strip()}\n</summary>")
    data_list = combine_content_in_list([head_list, [content], tail_list])
    write_list(data_list, html_name, "w", None, False)
    return None
